import tempfile
from os import path, listdir
from pathlib import Path
from typing import List, Dict, cast

import pytest

from genomics_data_index.storage.io.mutation.NucleotideSampleData import NucleotideSampleData
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.io.mutation.VcfVariantsReader import VcfVariantsReader
from genomics_data_index.storage.io.mutation.variants_processor.MultipleProcessVcfVariantsTableProcessor import \
    MultipleProcessVcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.mutation.variants_processor.SerialVcfVariantsTableProcessor import \
    SerialVcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.mutation.variants_processor.VcfVariantsTableProcessor import \
    VcfVariantsTableProcessorFactory
from genomics_data_index.storage.io.processor.SerialSampleFilesProcessor import SerialSampleFilesProcessor
from genomics_data_index.test.integration import data_dir
from genomics_data_index.test.integration import data_dir_empty
from genomics_data_index.test.integration import snpeff_sample_vcfs, snpeff_sarscov2_vcfs

serial_variants_processor_factory = SerialVcfVariantsTableProcessorFactory()


@pytest.fixture
def sample_dirs() -> List[Path]:
    return [data_dir / d for d in listdir(data_dir) if path.isdir(data_dir / d)]


@pytest.fixture
def sample_dirs_empty() -> List[Path]:
    return [data_dir_empty / d for d in listdir(data_dir_empty) if path.isdir(data_dir_empty / d)]


def vcf_and_mask_files(sample_dirs) -> Dict[str, Dict[str, Path]]:
    sample_vcf_map = {}
    sample_genomic_files_mask = {}
    for d in sample_dirs:
        sample_name = path.basename(d)
        vcf_file = Path(d, 'snps.vcf.gz')
        genomic_file_mask = Path(d, 'snps.aligned.fa')

        sample_vcf_map[sample_name] = vcf_file
        sample_genomic_files_mask[sample_name] = genomic_file_mask

    return {
        'vcfs': sample_vcf_map,
        'masks': sample_genomic_files_mask
    }


def variants_reader_internal(sample_dirs, variants_processor_factory: VcfVariantsTableProcessorFactory,
                             include_masked_regions: bool) -> VcfVariantsReader:
    tmp_dir = Path(tempfile.mkdtemp())
    vcf_masks = vcf_and_mask_files(sample_dirs)
    file_processor = SerialSampleFilesProcessor(tmp_dir)
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcf_masks['vcfs'],
                                                                     masked_genomic_files_map=vcf_masks[
                                                                         'masks'],
                                                                     variants_processor_factory=variants_processor_factory,
                                                                     sample_files_processor=file_processor,
                                                                     index_unknown_missing=include_masked_regions)
    processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
    processed_files = processed_data_package.get_sample_data()
    return VcfVariantsReader.create(processed_files, variants_processor_factory=variants_processor_factory,
                                    include_masked_regions=include_masked_regions)


@pytest.fixture
def variants_reader(sample_dirs) -> VcfVariantsReader:
    return variants_reader_internal(sample_dirs, variants_processor_factory=serial_variants_processor_factory,
                                    include_masked_regions=False)


@pytest.fixture
def variants_reader_empty(sample_dirs_empty) -> VcfVariantsReader:
    return variants_reader_internal(sample_dirs_empty, variants_processor_factory=serial_variants_processor_factory,
                                    include_masked_regions=False)


@pytest.fixture
def variants_reader_snpeff_annotations_single_sample() -> VcfVariantsReader:
    tmp_dir = Path(tempfile.mkdtemp())
    file_processor = SerialSampleFilesProcessor(tmp_dir)

    vcfs_map = {
        'SH10-014': snpeff_sample_vcfs['SH10-014']
    }

    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcfs_map,
                                                                     masked_genomic_files_map=None,
                                                                     sample_files_processor=file_processor)
    processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
    processed_files = processed_data_package.get_sample_data()
    return VcfVariantsReader.create(processed_files, variants_processor_factory=serial_variants_processor_factory,
                                    include_masked_regions=False)


@pytest.fixture
def variants_reader_snpeff_annotations_single_sample_sars_cov_2() -> VcfVariantsReader:
    tmp_dir = Path(tempfile.mkdtemp())
    file_processor = SerialSampleFilesProcessor(tmp_dir)

    vcfs_map = {
        'USA/CA-CDPH-3000143037/2021': snpeff_sarscov2_vcfs['USA/CA-CDPH-3000143037/2021']
    }

    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=vcfs_map,
                                                                     masked_genomic_files_map=None,
                                                                     sample_files_processor=file_processor)
    processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
    processed_files = processed_data_package.get_sample_data()
    return VcfVariantsReader.create(processed_files, variants_processor_factory=serial_variants_processor_factory,
                                    include_masked_regions=False)


@pytest.fixture
def variants_reader_snpeff_annotations_multiple_samples() -> VcfVariantsReader:
    tmp_dir = Path(tempfile.mkdtemp())
    file_processor = SerialSampleFilesProcessor(tmp_dir)
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs,
                                                                     masked_genomic_files_map=None,
                                                                     sample_files_processor=file_processor)
    processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
    processed_files = processed_data_package.get_sample_data()
    return VcfVariantsReader.create(processed_files, variants_processor_factory=serial_variants_processor_factory,
                                    include_masked_regions=False)


@pytest.fixture
def variants_reader_empty_masks(sample_dirs) -> VcfVariantsReader:
    sample_vcf_map = {}
    for d in sample_dirs:
        sample_name = path.basename(d)
        vcf_file = Path(d, 'snps.vcf.gz')

        sample_vcf_map[sample_name] = vcf_file

    tmp_dir = Path(tempfile.mkdtemp())
    file_processor = SerialSampleFilesProcessor(tmp_dir)
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=sample_vcf_map,
                                                                     sample_files_processor=file_processor)
    processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
    processed_files = processed_data_package.get_sample_data()
    return VcfVariantsReader.create(processed_files, variants_processor_factory=serial_variants_processor_factory,
                                    include_masked_regions=False)


@pytest.fixture
def variants_reader_default_no_data() -> VcfVariantsReader:
    return VcfVariantsReader(sample_files_map={}, variants_processor_factory=serial_variants_processor_factory)


def variants_reader_from_snippy_internal(sample_dirs, variants_processor_factory: VcfVariantsTableProcessorFactory,
                                         include_masked_regions: bool) -> VcfVariantsReader:
    tmp_dir = Path(tempfile.mkdtemp())
    file_processor = SerialSampleFilesProcessor(tmp_dir)
    data_package = NucleotideSampleDataPackage.create_from_snippy(sample_dirs=sample_dirs,
                                                                  sample_files_processor=file_processor)
    processed_data_package = cast(NucleotideSampleDataPackage, data_package.process_all_data())
    processed_files = processed_data_package.get_sample_data()
    return VcfVariantsReader.create(processed_files, variants_processor_factory=variants_processor_factory,
                                    include_masked_regions=include_masked_regions)


@pytest.fixture
def variants_reader_from_snippy(sample_dirs) -> VcfVariantsReader:
    return variants_reader_from_snippy_internal(sample_dirs,
                                                variants_processor_factory=serial_variants_processor_factory,
                                                include_masked_regions=False)


@pytest.fixture
def variants_reader_from_snippy_masked(sample_dirs) -> VcfVariantsReader:
    return variants_reader_from_snippy_internal(sample_dirs,
                                                variants_processor_factory=serial_variants_processor_factory,
                                                include_masked_regions=True)


@pytest.fixture
def variants_reader_from_snippy_parallel(sample_dirs) -> VcfVariantsReader:
    processing_factory = MultipleProcessVcfVariantsTableProcessorFactory(ncores=2)
    return variants_reader_from_snippy_internal(sample_dirs, variants_processor_factory=processing_factory,
                                                include_masked_regions=False)


@pytest.fixture
def variants_reader_from_snippy_masked_parallel(sample_dirs) -> VcfVariantsReader:
    processing_factory = MultipleProcessVcfVariantsTableProcessorFactory(ncores=2)
    return variants_reader_from_snippy_internal(sample_dirs, variants_processor_factory=processing_factory,
                                                include_masked_regions=True)


@pytest.fixture
def variants_reader_from_snippy_masked_multicore(sample_dirs) -> VcfVariantsReader:
    return variants_reader_from_snippy_internal(sample_dirs,
                                                variants_processor_factory=serial_variants_processor_factory,
                                                include_masked_regions=True)


@pytest.fixture
def variants_reader_snpeff() -> VcfVariantsReader:
    tmp_dir = Path(tempfile.mkdtemp())
    file_processor = SerialSampleFilesProcessor(tmp_dir)
    data_package = NucleotideSampleDataPackage.create_from_vcf_masks(sample_vcf_map=snpeff_sample_vcfs,
                                                                     sample_files_processor=file_processor)
    processed_files = cast(Dict[str, NucleotideSampleData], data_package.process_all_data())
    return VcfVariantsReader.create(processed_files, variants_processor_factory=serial_variants_processor_factory,
                                    include_masked_regions=False)


def test_get_variants_table(variants_reader):
    df = variants_reader.get_features_table()

    assert 129 == len(df), 'Data has incorrect length'
    assert {'SampleA', 'SampleB', 'SampleC'} == set(df['SAMPLE'].tolist()), 'Incorrect sample names'

    assert ['SNP'] == df.loc[(df['SAMPLE'] == 'SampleA') & (df['POS'] == 293), 'TYPE'].tolist()
    assert ['INDEL'] == df.loc[(df['SAMPLE'] == 'SampleA') & (df['POS'] == 302), 'TYPE'].tolist()
    assert ['INDEL'] == df.loc[(df['SAMPLE'] == 'SampleA') & (df['POS'] == 324), 'TYPE'].tolist()
    assert ['INDEL'] == df.loc[(df['SAMPLE'] == 'SampleA') & (df['POS'] == 374), 'TYPE'].tolist()
    assert ['OTHER'] == df.loc[(df['SAMPLE'] == 'SampleA') & (df['POS'] == 461), 'TYPE'].tolist()
    assert ['OTHER'] == df.loc[(df['SAMPLE'] == 'SampleB') & (df['POS'] == 1325), 'TYPE'].tolist()
    assert ['OTHER'] == df.loc[(df['SAMPLE'] == 'SampleC') & (df['POS'] == 1984), 'TYPE'].tolist()


def test_get_samples_list(variants_reader):
    assert {'SampleA', 'SampleB', 'SampleC'} == set(variants_reader.samples_list())


def test_get_variants_table_empty(variants_reader_empty):
    df = variants_reader_empty.get_features_table()

    assert 0 == len(df), 'Data has incorrect length'


def test_get_or_create_feature_file(variants_reader):
    file = variants_reader.get_or_create_feature_file('SampleA')
    assert file.exists()


def test_snippy_get_variants_table(variants_reader_from_snippy):
    df = variants_reader_from_snippy.get_features_table()

    assert 129 == len(df), 'Data has incorrect length'
    assert {'SampleA', 'SampleB', 'SampleC'} == set(df['SAMPLE'].tolist()), 'Incorrect sample names'


def test_snippy_get_variants_table_parallel(variants_reader_from_snippy_parallel):
    df = variants_reader_from_snippy_parallel.get_features_table()

    assert 129 == len(df), 'Data has incorrect length'
    assert {'SampleA', 'SampleB', 'SampleC'} == set(df['SAMPLE'].tolist()), 'Incorrect sample names'


def test_snippy_get_variants_table_masked(variants_reader_from_snippy_masked):
    df = variants_reader_from_snippy_masked.get_features_table()

    assert 1079 == len(df), 'Data has incorrect length'
    assert {'SampleA', 'SampleB', 'SampleC'} == set(df['SAMPLE'].tolist()), 'Incorrect sample names'

    # Missing/unknown
    assert 414 == len(df[(df['SAMPLE'] == 'SampleA') & (df['TYPE'] == 'UNKNOWN_MISSING')])
    assert 234 == len(df[(df['SAMPLE'] == 'SampleB') & (df['TYPE'] == 'UNKNOWN_MISSING')])
    assert 302 == len(df[(df['SAMPLE'] == 'SampleC') & (df['TYPE'] == 'UNKNOWN_MISSING')])

    # Variants
    assert 46 == len(df[(df['SAMPLE'] == 'SampleA') & (df['TYPE'] != 'UNKNOWN_MISSING')])
    assert 50 == len(df[(df['SAMPLE'] == 'SampleB') & (df['TYPE'] != 'UNKNOWN_MISSING')])
    assert 33 == len(df[(df['SAMPLE'] == 'SampleC') & (df['TYPE'] != 'UNKNOWN_MISSING')])


def test_snippy_get_variants_table_masked_parallel(variants_reader_from_snippy_masked_parallel):
    df = variants_reader_from_snippy_masked_parallel.get_features_table()

    assert 1079 == len(df), 'Data has incorrect length'
    assert {'SampleA', 'SampleB', 'SampleC'} == set(df['SAMPLE'].tolist()), 'Incorrect sample names'

    # Missing/unknown
    assert 414 == len(df[(df['SAMPLE'] == 'SampleA') & (df['TYPE'] == 'UNKNOWN_MISSING')])
    assert 234 == len(df[(df['SAMPLE'] == 'SampleB') & (df['TYPE'] == 'UNKNOWN_MISSING')])
    assert 302 == len(df[(df['SAMPLE'] == 'SampleC') & (df['TYPE'] == 'UNKNOWN_MISSING')])

    # Variants
    assert 46 == len(df[(df['SAMPLE'] == 'SampleA') & (df['TYPE'] != 'UNKNOWN_MISSING')])
    assert 50 == len(df[(df['SAMPLE'] == 'SampleB') & (df['TYPE'] != 'UNKNOWN_MISSING')])
    assert 33 == len(df[(df['SAMPLE'] == 'SampleC') & (df['TYPE'] != 'UNKNOWN_MISSING')])


def test_snippy_get_variants_table_masked_multicore(variants_reader_from_snippy_masked_multicore):
    df = variants_reader_from_snippy_masked_multicore.get_features_table()

    assert 1079 == len(df), 'Data has incorrect length'
    assert {'SampleA', 'SampleB', 'SampleC'} == set(df['SAMPLE'].tolist()), 'Incorrect sample names'

    # Missing/unknown
    assert 414 == len(df[(df['SAMPLE'] == 'SampleA') & (df['TYPE'] == 'UNKNOWN_MISSING')])
    assert 234 == len(df[(df['SAMPLE'] == 'SampleB') & (df['TYPE'] == 'UNKNOWN_MISSING')])
    assert 302 == len(df[(df['SAMPLE'] == 'SampleC') & (df['TYPE'] == 'UNKNOWN_MISSING')])

    # Variants
    assert 46 == len(df[(df['SAMPLE'] == 'SampleA') & (df['TYPE'] != 'UNKNOWN_MISSING')])
    assert 50 == len(df[(df['SAMPLE'] == 'SampleB') & (df['TYPE'] != 'UNKNOWN_MISSING')])
    assert 33 == len(df[(df['SAMPLE'] == 'SampleC') & (df['TYPE'] != 'UNKNOWN_MISSING')])


def test_snippy_get_genomic_masks(variants_reader_from_snippy):
    mask = variants_reader_from_snippy.get_genomic_masked_region('SampleA')
    assert 414 == len(mask)
    assert {'reference'} == mask.sequence_names()

    mask = variants_reader_from_snippy.get_genomic_masked_region('SampleB')
    assert 234 == len(mask)
    assert {'reference'} == mask.sequence_names()

    mask = variants_reader_from_snippy.get_genomic_masked_region('SampleC')
    assert 302 == len(mask)
    assert {'reference'} == mask.sequence_names()


def test_snippy_get_samples_list(variants_reader_from_snippy):
    assert {'SampleA', 'SampleB', 'SampleC'} == set(variants_reader_from_snippy.samples_list())


def test_snippy_get_samples_list_two_files():
    sample_dirs = [data_dir / 'SampleA', data_dir / 'SampleB']
    reader = variants_reader_from_snippy_internal(sample_dirs,
                                                  variants_processor_factory=serial_variants_processor_factory,
                                                  include_masked_regions=False)

    assert {'SampleA', 'SampleB'} == set(reader.samples_list())


def test_snippy_read_empty_vcf(sample_dirs_empty):
    reader = variants_reader_from_snippy_internal(sample_dirs_empty,
                                                  variants_processor_factory=serial_variants_processor_factory,
                                                  include_masked_regions=False)
    df = reader.get_features_table()

    assert 0 == len(df), 'Data has incorrect length'


def test_snippy_read_empty_vcf_include_masked_regions(sample_dirs_empty):
    reader = variants_reader_from_snippy_internal(sample_dirs_empty,
                                                  variants_processor_factory=serial_variants_processor_factory,
                                                  include_masked_regions=True)
    df = reader.get_features_table()

    assert 437 == len(df), 'Data has incorrect length'
    assert {'UNKNOWN_MISSING'} == set(df['TYPE'])
    assert {'?'} == set(df['ALT'])


def test_snippy_read_empty_vcf_multicore_include_masked_regions(sample_dirs_empty):
    reader = variants_reader_from_snippy_internal(sample_dirs_empty,
                                                  variants_processor_factory=serial_variants_processor_factory,
                                                  include_masked_regions=True)
    df = reader.get_features_table()

    assert 437 == len(df), 'Data has incorrect length'
    assert {'UNKNOWN_MISSING'} == set(df['TYPE'])
    assert {'?'} == set(df['ALT'])


def test_get_variants_table_snpeff_annotations_single_sample(
        variants_reader_snpeff_annotations_single_sample: VcfVariantsReader):
    vr = variants_reader_snpeff_annotations_single_sample
    vcf_df = vr.get_features_table()

    assert 139 == len(vcf_df)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        vcf_df.columns)

    # Get rid of 'FILE' since these are processed files
    vcf_df = vcf_df.drop(columns='FILE')

    # # missense variant
    vcf_df_varA = vcf_df[vcf_df['POS'] == 140658]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 140658, 'C', 'A', 'SNP', 'NC_011083:140658:C:A',
            'A', 'missense_variant', 'MODERATE', 'murF', 'SEHA_RS01180', 'transcript', 'protein_coding',
            'c.497C>A', 'p.Ala166Glu'] == vcf_df_varA.iloc[0].tolist()

    # Synonymous variant
    vcf_df_varA = vcf_df[vcf_df['POS'] == 723772]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 723772, 'T', 'C', 'SNP', 'NC_011083:723772:T:C',
            'C', 'synonymous_variant', 'LOW', 'SEHA_RS03990', 'SEHA_RS03990', 'transcript', 'protein_coding',
            'c.1638T>C', 'p.Cys546Cys'] == vcf_df_varA.iloc[0].tolist()

    # Intergenic variant (snp)
    vcf_df_varA = vcf_df[vcf_df['POS'] == 1031571]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 1031571, 'T', 'C', 'SNP', 'NC_011083:1031571:T:C',
            'C', 'intergenic_region', 'MODIFIER', 'SEHA_RS05505-SEHA_RS05515', 'SEHA_RS05505-SEHA_RS05515',
            'intergenic_region', 'NA',
            'n.1031571T>C', 'NA'] == vcf_df_varA.iloc[0].fillna('NA').tolist()

    # Intergenic variant (del)
    vcf_df_varA = vcf_df[vcf_df['POS'] == 4555461]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 4555461, 'T', 'TC', 'INDEL', 'NC_011083:4555461:T:TC',
            'TC', 'intergenic_region', 'MODIFIER', 'SEHA_RS22510-SEHA_RS26685', 'SEHA_RS22510-SEHA_RS26685',
            'intergenic_region', 'NA',
            'n.4555461_4555462insC', 'NA'] == vcf_df_varA.iloc[0].fillna('NA').tolist()

    # frameshift variant
    vcf_df_varA = vcf_df[vcf_df['POS'] == 1770751]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 1770751, 'CA', 'C', 'INDEL', 'NC_011083:1770751:CA:C',
            'C', 'frameshift_variant', 'HIGH', 'SEHA_RS09270', 'SEHA_RS09270',
            'transcript', 'protein_coding',
            'c.335delT', 'p.Leu112fs'] == vcf_df_varA.iloc[0].fillna('NA').tolist()

    # stop gained variant
    vcf_df_varA = vcf_df[vcf_df['POS'] == 4882099]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 4882099, 'C', 'T', 'SNP', 'NC_011083:4882099:C:T',
            'T', 'stop_gained', 'HIGH', 'SEHA_RS24155', 'SEHA_RS24155',
            'transcript', 'protein_coding',
            'c.1287G>A', 'p.Trp429*'] == vcf_df_varA.iloc[0].fillna('NA').tolist()

    # stop lost variant
    vcf_df_varA = vcf_df[vcf_df['POS'] == 4824790]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 4824790, 'T', 'C', 'SNP', 'NC_011083:4824790:T:C',
            'C', 'stop_lost', 'HIGH', 'SEHA_RS23860', 'SEHA_RS23860',
            'transcript', 'protein_coding',
            'c.1378T>C', 'p.Ter460Glnext*?'] == vcf_df_varA.iloc[0].fillna('NA').tolist()

    # inframe deletion
    vcf_df_varA = vcf_df[vcf_df['POS'] == 4465400]
    assert 1 == len(vcf_df_varA)
    assert ['SH10-014', 'NC_011083', 4465400, 'GGCCGAA', 'G', 'INDEL', 'NC_011083:4465400:GGCCGAA:G',
            'G', 'conservative_inframe_deletion', 'MODERATE', 'tyrB', 'SEHA_RS22180',
            'transcript', 'protein_coding',
            'c.157_162delGAAGCC', 'p.Glu53_Ala54del'] == vcf_df_varA.iloc[0].fillna(
        'NA').tolist()


def test_get_variants_table_snpeff_annotations_multiple_samples(
        variants_reader_snpeff_annotations_multiple_samples: VcfVariantsReader):
    vr = variants_reader_snpeff_annotations_multiple_samples
    vcf_df = vr.get_features_table()

    assert 361 == len(vcf_df)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        vcf_df.columns)

    # Get rid of 'FILE' since these are processed files
    vcf_df = vcf_df.drop(columns='FILE')

    # disruptive inframe deletion
    vcf_df_varA = vcf_df[vcf_df['POS'] == 3167187]
    assert 3 == len(vcf_df_varA)
    vcf_df_varA = vcf_df_varA.set_index('SAMPLE')
    assert ['NC_011083', 3167187, 'AACCACGACCACGACCACGACCACGACCACGACCACGACCACGACCACG', 'A', 'INDEL',
            'NC_011083:3167187:AACCACGACCACGACCACGACCACGACCACGACCACGACCACGACCACG:A',
            'A', 'disruptive_inframe_deletion', 'MODERATE', 'SEHA_RS15905', 'SEHA_RS15905', 'transcript',
            'protein_coding',
            'c.417_464delCGACCACGACCACGACCACGACCACGACCACGACCACGACCACGACCA', 'p.Asp140_His155del'] == vcf_df_varA.loc[
               'SH10-014'].tolist()
    assert ['NC_011083', 3167187, 'AACCACGACCACGACCACGACCACGACCACGACCACG', 'A', 'INDEL',
            'NC_011083:3167187:AACCACGACCACGACCACGACCACGACCACGACCACG:A',
            'A', 'disruptive_inframe_deletion', 'MODERATE', 'SEHA_RS15905', 'SEHA_RS15905', 'transcript',
            'protein_coding',
            'c.429_464delCGACCACGACCACGACCACGACCACGACCACGACCA', 'p.Asp144_His155del'] == vcf_df_varA.loc[
               'SH14-001'].tolist()
    assert ['NC_011083', 3167187, 'AACCACGACCACGACCACGACCACGACCACGACCACG', 'A', 'INDEL',
            'NC_011083:3167187:AACCACGACCACGACCACGACCACGACCACGACCACG:A',
            'A', 'disruptive_inframe_deletion', 'MODERATE', 'SEHA_RS15905', 'SEHA_RS15905', 'transcript',
            'protein_coding',
            'c.429_464delCGACCACGACCACGACCACGACCACGACCACGACCA', 'p.Asp144_His155del'] == vcf_df_varA.loc[
               'SH14-014'].tolist()


def test_get_variants_table_snpeff_annotations_sars_cov_2(
        variants_reader_snpeff_annotations_single_sample_sars_cov_2: VcfVariantsReader):
    vr = variants_reader_snpeff_annotations_single_sample_sars_cov_2
    vcf_df = vr.get_features_table()

    assert 31 == len(vcf_df)
    assert ['SAMPLE', 'CHROM', 'POS', 'REF', 'ALT', 'TYPE', 'FILE', 'VARIANT_ID',
            'ANN.Allele', 'ANN.Annotation', 'ANN.Annotation_Impact', 'ANN.Gene_Name', 'ANN.Gene_ID',
            'ANN.Feature_Type', 'ANN.Transcript_BioType', 'ANN.HGVS.c', 'ANN.HGVS.p'] == list(
        vcf_df.columns)

    # Get rid of 'FILE' since these are processed files
    vcf_df = vcf_df.drop(columns='FILE')

    vcf_df['ANN.Annotation'] = vcf_df['ANN.Annotation'].astype(str)
    vcf_df['ANN.Annotation_Impact'] = vcf_df['ANN.Annotation_Impact'].astype(str)

    # ORF1ab (ORF1a region)
    sample_sarscov2_1_var = vcf_df[vcf_df['POS'] == 3948]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 3948, 'A', 'G', 'SNP',
            'NC_045512.2:3948:A:G', 'G', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript',
            'protein_coding', 'c.3683A>G', 'p.D1228G'] == sample_sarscov2_1_var.iloc[0].tolist()

    # ORF1ab (ORF1a region)
    sample_sarscov2_1_var = vcf_df[vcf_df['POS'] == 3037]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 3037, 'C', 'T', 'SNP',
            'NC_045512.2:3037:C:T', 'T', 'synonymous_variant', 'LOW', 'ORF1ab', 'GU280_gp01', 'transcript',
            'protein_coding', 'c.2772C>T', 'p.F924F'] == sample_sarscov2_1_var.iloc[0].tolist()

    # ORF1ab (ORF1b region)
    sample_sarscov2_1_var = vcf_df[vcf_df['POS'] == 19220]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 19220, 'C', 'T', 'SNP',
            'NC_045512.2:19220:C:T', 'T', 'missense_variant', 'MODERATE', 'ORF1ab', 'GU280_gp01', 'transcript',
            'protein_coding', 'c.18956C>T', 'p.A6319V'] == sample_sarscov2_1_var.iloc[0].tolist()

    # S
    sample_sarscov2_1_var = vcf_df[vcf_df['POS'] == 22917]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 22917, 'T', 'G', 'SNP',
            'NC_045512.2:22917:T:G', 'G', 'missense_variant', 'MODERATE', 'S', 'GU280_gp02', 'transcript',
            'protein_coding', 'c.1355T>G', 'p.L452R'] == sample_sarscov2_1_var.iloc[0].tolist()

    # ORF7b
    sample_sarscov2_1_var = vcf_df[vcf_df['POS'] == 27874]
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 27874, 'C', 'T', 'SNP',
            'NC_045512.2:27874:C:T', 'T', 'missense_variant', 'MODERATE', 'ORF7b', 'GU280_gp08', 'transcript',
            'protein_coding', 'c.119C>T', 'p.T40I'] == sample_sarscov2_1_var.iloc[0].tolist()

    # intergenic
    sample_sarscov2_1_var = vcf_df[vcf_df['POS'] == 210].fillna('<NA>')
    assert 1 == len(sample_sarscov2_1_var)
    assert ['USA/CA-CDPH-3000143037/2021', 'NC_045512.2', 210, 'G', 'T', 'SNP',
            'NC_045512.2:210:G:T', 'T', 'intergenic_region', 'MODIFIER', 'CHR_START-ORF1ab', 'CHR_START-GU280_gp01',
            'intergenic_region', '<NA>', 'n.210G>T', '<NA>'] == sample_sarscov2_1_var.iloc[0].tolist()



from __future__ import annotations

import logging
import os
from pathlib import Path
from typing import Union

from genomics_data_index.configuration.ConfigFile import ConfigFile
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection

logger = logging.getLogger(__name__)


class ProjectConfigurationError(Exception):

    def __init__(self, msg: str):
        super().__init__(msg)


class Project:
    DATA_DIR = '.gdi-data'
    DATABASE_FILE = 'gdi-db.sqlite'
    CONFIG_FILE = 'gdi-config.yaml'

    def __init__(self, root_dir: Union[Path, str]):
        if isinstance(root_dir, str):
            root_dir = Path(root_dir)

        self._root_dir = root_dir
        self._config_file = root_dir / self.CONFIG_FILE

        if not self._config_file.exists():
            raise ProjectConfigurationError(f'Config file [{self._config_file}] does not exist. '
                                            f'Please verify that project directory [{root_dir}] is valid.')

        self._config = ConfigFile.read_config(self._config_file)
        self._database_connection = self.get_database_connection_str()
        self._database_dir = self.get_database_dir()

    def get_root_dir(self) -> Path:
        return self._root_dir

    def get_database_connection_str(self) -> str:
        if self._config.database_connection is None:
            if self._config.sqlite_database is None:
                raise ProjectConfigurationError(f'No valid configured database in config file {self._config_file}')
            else:
                # Provide support for defining sqlite file relative to project directory or as an absolute string
                if self._config.sqlite_database.is_absolute():
                    sqlite_path = self._config.sqlite_database
                else:
                    sqlite_path = self._root_dir / self._config.sqlite_database

            database_connection = f'sqlite:///{sqlite_path}'
        elif self._config.database_connection is None:
            raise ProjectConfigurationError(f'No valid configured database in config file {self._config_file}')
        else:
            database_connection = self._config.database_connection

        return database_connection

    def get_database_dir(self) -> Path:
        if self._config.database_dir is None:
            raise ProjectConfigurationError(
                f'database_dir is not configured properly in config file {self._config_file}')

        if self._config.database_dir.is_absolute():
            database_dir = self._config.database_dir
        else:
            database_dir = self._root_dir / self._config.database_dir

        if not database_dir.exists():
            raise ProjectConfigurationError(f'database_dir=[{database_dir}] does not exist. '
                                            f'Please verify that project directory [{self._root_dir}] is valid.')
        else:
            return database_dir

    def create_connection(self) -> DataIndexConnection:
        return DataIndexConnection.connect(database_connection=self._database_connection,
                                           database_dir=self._database_dir)

    @classmethod
    def initialize_project(cls, project_dir: Path) -> Project:
        if not project_dir.exists():
            os.mkdir(project_dir)

        config_file = project_dir / cls.CONFIG_FILE
        data_dir = project_dir / cls.DATA_DIR

        if data_dir.exists() or config_file.exists():
            raise Exception(f'Project directory [{project_dir}] has already been initialized')

        os.mkdir(data_dir)

        config = ConfigFile()
        config.sqlite_database = cls.DATA_DIR + os.path.sep + cls.DATABASE_FILE
        config.database_dir = cls.DATA_DIR

        config.write(project_dir / cls.CONFIG_FILE)

        return Project(project_dir)



import abc
from typing import Union, List

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.kind.isa.DelegateIsaKind import DelegateIsaKind
from genomics_data_index.storage.SampleSet import SampleSet


class SamplesTypingIsaKind(DelegateIsaKind, abc.ABC):

    def __init__(self):
        super().__init__()

    def isa(self, data: Union[str, List[str], SamplesQuery, SampleSet], query: SamplesQuery) -> SamplesQuery:
        if isinstance(data, str):
            return self.isa_type(data, query)
        else:
            raise Exception(f'Can only apply {self.__class__} to data of type string. Got data={data}')

    @property
    @abc.abstractmethod
    def name(self) -> str:
        pass

    @property
    @abc.abstractmethod
    def version(self) -> str:
        pass

    @abc.abstractmethod
    def isa_type(self, data: str, query: SamplesQuery) -> SamplesQuery:
        pass



from pathlib import Path

from genomics_data_index.storage.io.mutation.SequenceFile import SequenceFile


# Method used to shorten creating sequence files with a given name
def sf_name(name: str) -> SequenceFile:
    return SequenceFile(Path(name))


def test_get_genome_name():
    assert 'file' == sf_name('file.gbk').get_genome_name()
    assert 'file_1' == sf_name('file_1.gbk').get_genome_name()
    assert 'file' == sf_name('file.gbk.gz').get_genome_name()
    assert 'file' == sf_name('file.fasta.gz').get_genome_name()

    # Should only exclude if it's reads
    assert 'file_1' == sf_name('file_1.gbk').get_genome_name(exclude_paired_end_indicators=True)

    assert 'file' == sf_name('file.fastq.gz').get_genome_name()
    assert 'file_A' == sf_name('file_A.fastq').get_genome_name()
    assert 'file_A_1' == sf_name('file_A_1.fastq').get_genome_name()
    assert 'file_A' == sf_name('file_A_1.fastq').get_genome_name(exclude_paired_end_indicators=True)
    assert 'file_A_R1' == sf_name('file_A_R1.fastq').get_genome_name()
    assert 'file_A' == sf_name('file_A_R1.fastq').get_genome_name(exclude_paired_end_indicators=True)
    assert 'file_A_R2' == sf_name('file_A_R2.fastq').get_genome_name()
    assert 'file_A' == sf_name('file_A_R2.fastq').get_genome_name(exclude_paired_end_indicators=True)
    assert 'file_A_R1_xyz' == sf_name('file_A_R1_xyz.fastq').get_genome_name()
    assert 'file_A' == sf_name('file_A_R1_xyz.fastq').get_genome_name(exclude_paired_end_indicators=True)
    assert 'file_A_R2_xyz' == sf_name('file_A_R2_xyz.fq.gz').get_genome_name()
    assert 'file_A' == sf_name('file_A_R2_xyz.fq.gz').get_genome_name(exclude_paired_end_indicators=True)


def test_is_genbank():
    assert SequenceFile(Path('file.gbk')).is_genbank()
    assert SequenceFile(Path('file.gb')).is_genbank()
    assert SequenceFile(Path('file.gbk.gz')).is_genbank()
    assert SequenceFile(Path('file.gb.gz')).is_genbank()

    assert not SequenceFile(Path('file.fasta')).is_genbank()
    assert not SequenceFile(Path('file.fasta.gz')).is_genbank()


def test_is_fasta():
    assert SequenceFile(Path('file.fasta')).is_fasta()
    assert SequenceFile(Path('file.fna')).is_fasta()
    assert SequenceFile(Path('file.fa')).is_fasta()
    assert SequenceFile(Path('file.fasta.gz')).is_fasta()
    assert SequenceFile(Path('file.fna.gz')).is_fasta()
    assert SequenceFile(Path('file.fa.gz')).is_fasta()

    assert not SequenceFile(Path('file.gb')).is_fasta()
    assert not SequenceFile(Path('file.fastq.gz')).is_fasta()


def test_is_assembly():
    assert SequenceFile(Path('file.fasta')).is_assembly()
    assert SequenceFile(Path('file.fna')).is_assembly()
    assert SequenceFile(Path('file.gb')).is_assembly()
    assert SequenceFile(Path('file.gbk.gz')).is_assembly()
    assert SequenceFile(Path('file.fna.gz')).is_assembly()
    assert SequenceFile(Path('file.fasta.gz')).is_assembly()

    assert not SequenceFile(Path('file.fastq')).is_assembly()
    assert not SequenceFile(Path('file.fastq.gz')).is_assembly()


def test_is_reads():
    assert SequenceFile(Path('file.fastq')).is_reads()
    assert SequenceFile(Path('file.fq')).is_reads()
    assert SequenceFile(Path('file.fastq.gz')).is_reads()
    assert SequenceFile(Path('file.fq.gz')).is_reads()

    assert not SequenceFile(Path('file.fasta')).is_reads()
    assert not SequenceFile(Path('file.gbk.gz')).is_reads()


def test_name_differences():
    assert [] == sf_name('file.fastq').name_differences(sf_name('file.fastq'))
    assert [] == sf_name('').name_differences(sf_name(''))
    assert [] == sf_name('.').name_differences(sf_name('.'))

    assert ['1'] == sf_name('file1.fastq').name_differences(sf_name('file2.fastq'))
    assert ['2'] == sf_name('file2.fastq').name_differences(sf_name('file1.fastq'))

    assert ['1'] == sf_name('file_1.fastq').name_differences(sf_name('file_2.fastq'))
    assert ['2'] == sf_name('file_2.fastq').name_differences(sf_name('file_1.fastq'))

    assert ['1'] == sf_name('file_xyz_R1.fastq').name_differences(sf_name('file_xyz_R2.fastq'))
    assert ['2'] == sf_name('file_xyz_R2.fastq').name_differences(sf_name('file_xyz_R1.fastq'))

    assert ['a', '1'] == sf_name('file_a_xyz_R1.fastq').name_differences(sf_name('file_b_xyz_R2.fastq'))
    assert ['b', '2'] == sf_name('file_b_xyz_R2.fastq').name_differences(sf_name('file_a_xyz_R1.fastq'))

    # Difference at end
    assert ['gz'] == sf_name('file.fastq.gz').name_differences(sf_name('file.fastq.by'))

    # Difference at beginning
    assert ['p.'] == sf_name('p.file.fastq.gz').name_differences(sf_name('x_file.fastq.gz'))

    # Many differences in beginning, middle, and end
    assert ['p.', '1', 'gz'] == sf_name('p.file1.fastq.gz').name_differences(sf_name('x_file2.fastq.by'))
    assert ['x_', '2', 'by'] == sf_name('x_file2.fastq.by').name_differences(sf_name('p.file1.fastq.gz'))



from typing import cast

from genomics_data_index.api.query.GenomicsDataIndex import GenomicsDataIndex
from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl import ExperimentalTreeSamplesQuery
from genomics_data_index.configuration.connector import DataIndexConnection


# wrapper methods to simplify writing tests
def query(connection: DataIndexConnection, **kwargs) -> SamplesQuery:
    return GenomicsDataIndex(connection=connection).samples_query(**kwargs)


def test_query_pruning_tree(loaded_database_connection: DataIndexConnection):
    query_result = query(loaded_database_connection).build_tree(
        kind='mutation_experimental', scope='genome',
        include_reference=True, extra_params='--seed 42 -m GTR').reset_universe()
    assert 3 == len(query_result)
    assert 3 == len(query_result.universe_set)

    query_result = cast(ExperimentalTreeSamplesQuery, query_result)

    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(
        query_result.tree.get_leaf_names())
    assert {'SampleA', 'genome'} == set(
        query_result.isa('SampleA').tree.get_leaf_names())
    assert {'SampleA', 'SampleB', 'genome'} == set(
        query_result.isin(['SampleA', 'SampleB']).tree.get_leaf_names())
    assert {'SampleB', 'SampleC', 'genome'} == set(
        query_result.isin(['SampleB', 'SampleC']).tree.get_leaf_names())

    # Test complement
    assert {'SampleB', 'SampleC', 'genome'} == set(
        query_result.isin('SampleA').complement().tree.get_leaf_names())

    # Double complement
    assert {'SampleB', 'genome'} == set(
        query_result.isin('SampleB').complement().complement().tree.get_leaf_names())

    # Test reset universe
    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(
        query_result.isa('SampleA').universe_tree.get_leaf_names())
    assert {'SampleA', 'genome'} == set(
        query_result.isa('SampleA').reset_universe().universe_tree.get_leaf_names())

    # Test reset universe 2
    assert {'SampleA', 'SampleB', 'SampleC', 'genome'} == set(
        query_result.isa(['SampleB', 'SampleC']).universe_tree.get_leaf_names())
    assert {'SampleB', 'SampleC', 'genome'} == set(
        query_result.isa(['SampleB', 'SampleC']).reset_universe().universe_tree.get_leaf_names())

    # Test reset universe and complement
    assert {'genome'} == set(
        query_result.isin('SampleA').reset_universe().complement().tree.get_leaf_names())



import pandas as pd
import pytest

from genomics_data_index.storage.io.mlst.MLSTSampleData import MLSTSampleData


@pytest.fixture
def mlst_sample_data() -> MLSTSampleData:
    allele_data = pd.Series({'l1': '1', 'l2': '2'}, name='Allele', dtype=str)
    return MLSTSampleData(sample_name='SampleA', scheme='lmonocytogenes',
                          mlst_allele_data=allele_data)


def test_single_sample_data(mlst_sample_data: MLSTSampleData):
    assert 'lmonocytogenes' == mlst_sample_data.get_scheme()
    assert 'SampleA' == mlst_sample_data.sample_name
    assert mlst_sample_data.is_preprocessed()
    assert 2 == len(mlst_sample_data.get_alleles())



from __future__ import annotations

from typing import Callable, Set

from ete3 import Tree, TreeNode

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.MutationTreeSamplesQuery import MutationTreeSamplesQuery
from genomics_data_index.api.query.impl.WrappedSamplesQuery import WrappedSamplesQuery
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet


class ExperimentalTreeSamplesQuery(MutationTreeSamplesQuery):
    """
    This class contains experimental selection/query statements with respect to a tree
    that I have started implementing but haven't had time to properly test out yet. Eventually
    these may make it into the regular TreeSamplesQuery class.
    """

    def __init__(self, connection: DataIndexConnection, wrapped_query: SamplesQuery,
                 tree: Tree, universe_tree: Tree,
                 alignment_length: int, reference_name: str, reference_included: bool):
        """
        Builds a new ExperimentalTreeSamplesQuery from the given information. In most normal operations SamplesQuery objects
        are not created directly but are instead created from an :py:class:`genomics_data_index.api.GenomicsDataIndex`
        object or from operations applied to a SamplesQuery (e.g., build_tree() or join_tree()).

        :param connection: A connection to a database containing samples.
        :param wrapped_query: The SamplesQuery to wrap around and join to the passed tree.
        :param tree: The tree to join to this query.
        :param alignment_length: The length of the alignment of mutations used to generate the tree
                                 (used to convert substitutions/site distances to substitutions distances).
        :param universe_tree: A tree representing the entire universe of samples in this query.
        :param reference_name: The name of the reference genome in the tree.
        :param reference_included: True if the reference genome is one of the leaves of the tree, False otherwise.
        :return: A new ExperimentalTreeSamplesQuery object.
        """
        super().__init__(connection=connection, wrapped_query=wrapped_query,
                         tree=tree, alignment_length=alignment_length,
                         reference_name=reference_name,
                         reference_included=reference_included)
        self._universe_tree = universe_tree

    @property
    def universe_tree(self) -> Tree:
        return self._universe_tree

    def reset_universe(self, include_unknown: bool = True) -> SamplesQuery:
        wrapped_reset_universe = self._wrapped_query.reset_universe(include_unknown=include_unknown)
        universe_tree = self._tree_copy_prune(from_query=wrapped_reset_universe, preserve_branch_length=True)
        return ExperimentalTreeSamplesQuery(connection=self._query_connection,
                                            wrapped_query=wrapped_reset_universe,
                                            tree=universe_tree,
                                            universe_tree=universe_tree,
                                            alignment_length=self._alignment_length,
                                            reference_name=self.reference_name,
                                            reference_included=self.reference_included)

    def _tree_copy(self) -> Tree:
        return self._tree.copy(method='cpickle')

    def _universe_tree_copy(self) -> Tree:
        return self._universe_tree.copy(method='cpickle')

    def _tree_copy_prune(self, from_query: SamplesQuery = None, preserve_branch_length: bool = True) -> Tree:
        tree = self._universe_tree_copy()
        if from_query is not None:
            query = from_query
        else:
            query = self

        if self.reference_included:
            nodes_to_keep = query.tolist() + [self.reference_name]
        else:
            nodes_to_keep = query.tolist()
        tree.prune(nodes_to_keep, preserve_branch_length=preserve_branch_length)
        return tree

    def relabel_samples(self, rename_func: Callable[[str], str]) -> SamplesQuery:
        """
        Relabels samples/leaves in the tree based on the passed function.
        :param rename_func: A function used for relabeling. This function takes as input the leaf name
                            and returns a new name for this leaf.
        :return: A ExperimentalTreeSamplesQuery where the tree has been relabeled.
        """
        tree = self._tree_copy()
        for node in tree.traverse("postorder"):
            if node.is_leaf():
                node.name = rename_func(node.name)
        return self._create_from_tree_internal(tree)

    def prune(self, preserve_branch_length: bool = True) -> SamplesQuery:
        """
        Prunes the tree joined to this ExperimentalTreeSamplesQuery down to only the currently selected samples.
        :param preserve_branch_length: True if branch lengths between the subset of samples should be preserved, False otherwise.
        :return: A new ExperimentalTreeSamplesQuery with the tree pruned down to only the selected samples (and the reference
                 genome if it is one of the leaves of the tree).
        """
        tree = self._tree_copy_prune(preserve_branch_length=preserve_branch_length)
        return self._create_from_tree_internal(tree)

    def label_internal_nodes(self) -> SamplesQuery:
        """
        Relabels internal nodes based on the selected samples.
        :return: A new ExperimentalTreeSamplesQuery with the internal nodes of the tree labeled.
        """
        tree = self._tree_copy()
        self._label_internal_nodes_recursive(tree)
        return self._create_from_tree_internal(tree)

    def _label_internal_nodes_recursive(self, node: TreeNode) -> Set[str]:
        if node.is_leaf():
            return {node.name}
        else:
            children_names = set()
            children = node.get_children()
            for child in children:
                children_names.update(self._label_internal_nodes_recursive(child))

            node.name = f'{children_names}'
            return children_names

    def _wrap_create(self, wrapped_query: SamplesQuery, universe_set: SampleSet = None) -> WrappedSamplesQuery:
        tree = self._tree_copy_prune(from_query=wrapped_query, preserve_branch_length=True)
        return ExperimentalTreeSamplesQuery(connection=self._query_connection,
                                            wrapped_query=wrapped_query,
                                            tree=tree,
                                            universe_tree=self._universe_tree,
                                            alignment_length=self._alignment_length,
                                            reference_name=self.reference_name,
                                            reference_included=self.reference_included)

    def _create_from_tree_internal(self, tree: Tree) -> SamplesQuery:
        return ExperimentalTreeSamplesQuery(connection=self._query_connection,
                                            wrapped_query=self._wrapped_query,
                                            tree=tree,
                                            universe_tree=self._universe_tree,
                                            alignment_length=self._alignment_length,
                                            reference_name=self.reference_name,
                                            reference_included=self.reference_included)

    @classmethod
    def from_tree_query(self, query: MutationTreeSamplesQuery) -> ExperimentalTreeSamplesQuery:
        """
        Creates new ExperimentalTreeSamplesQuery to match the same information as the passed query.
        :param query: The query to copy from.
        :return: An equivalent experimental query.
        """
        return ExperimentalTreeSamplesQuery(connection=query._query_connection,
                                            wrapped_query=query._wrapped_query,
                                            tree=query.tree,
                                            universe_tree=query.tree,
                                            alignment_length=query._alignment_length,
                                            reference_name=query.reference_name,
                                            reference_included=query.reference_included)



import logging
import multiprocessing as mp
from pathlib import Path
from typing import Generator, List

from genomics_data_index.storage.io.SampleData import SampleData
from genomics_data_index.storage.io.SampleFilesProcessor import SampleFilesProcessor
from genomics_data_index.storage.util import TRACE_LEVEL

logger = logging.getLogger(__name__)


class MultipleProcessSampleFilesProcessor(SampleFilesProcessor):

    def __init__(self, preprocess_dir: Path, processing_cores: int, max_chunk_size=25):
        super().__init__()
        self._preprocess_dir = preprocess_dir
        self._processing_cores = processing_cores
        self._max_chunk_size = max_chunk_size

    def handle_single_file(self, sample_files: SampleData) -> SampleData:
        return sample_files.persist(self._preprocess_dir)

    def _get_chunk_size(self, number_samples: int):
        chunk_size = max(1, int(number_samples / self._processing_cores))
        return min(self._max_chunk_size, chunk_size)

    def process(self, sample_data: List[SampleData]) -> Generator[SampleData, None, None]:
        number_samples = len(sample_data)
        chunk_size = self._get_chunk_size(number_samples)
        logger.debug(f'Starting preprocessing {number_samples} samples '
                     f'with {self._processing_cores} cores and chunk size {chunk_size}')
        with mp.Pool(self._processing_cores) as pool:
            processed_sample_filed = pool.imap_unordered(self.handle_single_file,
                                                         sample_data,
                                                         chunk_size)
            for processed_file in processed_sample_filed:
                logger.log(TRACE_LEVEL, f'Finished {processed_file.sample_name}')
                yield processed_file

        logger.debug(f'Finished preprocessing {number_samples} samples')



#!/usr/bin/env python
# Prints the sequential positions of all intervals in a bed file (1-based coordinates)
# Usage: cat file.bed | print-positions-bed.py 

import sys

from pybedtools import BedTool

lines = ''.join(sys.stdin.readlines())
bed = BedTool(lines, from_string=True)

for i in bed:
    start = i.start + 1
    stop = i.stop + 1
    for pos in range(start, stop):
        print(f'{i.chrom}\t{pos}')



from __future__ import annotations

from typing import Union, List, Optional

from genomics_data_index.storage.model import NUCLEOTIDE_UNKNOWN
from genomics_data_index.storage.model.QueryFeature import QueryFeature
from genomics_data_index.storage.model.QueryFeatureMutation import QueryFeatureMutation


class QueryFeatureMutationSPDI(QueryFeatureMutation):

    def __init__(self, spdi: str):
        super().__init__()
        self._spdi = spdi

        seq, pos, ref, alt = self._spdi.split(self.SPLIT_CHAR)

        if seq == self.WILD:
            raise Exception(f'Cannot set seq to be wild ({self.WILD}): {spdi}')
        else:
            self._seq_name = seq

        if pos == self.WILD:
            raise Exception(f'Cannot set pos to be wild ({self.WILD}): {spdi}')
        else:
            self._pos = int(pos)

        if ref == self.WILD:
            raise Exception(f'Cannot set ref to be wild ({self.WILD}): {spdi}')
        elif ref.isdigit():
            self._ref = int(ref)
        else:
            self._ref = ref

        if alt is None:
            self._alt = self.WILD
        else:
            self._alt = alt

    @property
    def id_no_prefix(self) -> Optional[str]:
        return self._spdi

    @property
    def scope(self) -> str:
        return self._seq_name

    @property
    def sequence(self) -> str:
        return self.scope

    @property
    def position(self) -> int:
        return self._pos

    @property
    def start(self) -> int:
        return self.position

    @property
    def stop(self) -> int:
        if isinstance(self._ref, int):
            return self.position + self._ref
        else:
            return self.position + len(self.ref)

    @property
    def start0(self) -> int:
        return self.start - 1

    @property
    def stop0(self) -> int:
        return self.stop - 1

    @property
    def ref(self) -> Union[str, int]:
        return self._ref

    def has_deletion_sequence(self) -> bool:
        return isinstance(self.ref, str)

    def deletion_length(self) -> int:
        if self.has_deletion_sequence():
            return len(self.deletion)
        else:
            return self.deletion

    @property
    def deletion(self) -> Union[str, int]:
        return self.ref

    @property
    def alt(self) -> str:
        return self._alt

    @property
    def insertion(self) -> str:
        return self.alt

    def is_unknown(self) -> bool:
        return self.insertion == NUCLEOTIDE_UNKNOWN

    def to_unknown(self) -> QueryFeature:
        return QueryFeatureMutationSPDI(':'.join([
            self._seq_name,
            str(self._pos),
            str(self._ref),
            NUCLEOTIDE_UNKNOWN
        ]))

    def to_unknown_explode(self) -> List[QueryFeature]:
        if self.deletion_length() == 1:
            unknowns = [self.to_unknown()]
        elif self.has_deletion_sequence():
            unknowns = []
            for i, c in enumerate(self.ref):
                unknowns.append(
                    QueryFeatureMutationSPDI(f'{self.sequence}:{self.position + i}:{c}:{NUCLEOTIDE_UNKNOWN}'))
        else:
            unknowns = []
            for i in range(self.ref):
                unknowns.append(QueryFeatureMutationSPDI(f'{self.sequence}:{self.position + i}:1:{NUCLEOTIDE_UNKNOWN}'))
        return unknowns



import abc
from typing import Generator

from genomics_data_index.storage.io.SampleDataPackage import SampleDataPackage


class SampleDataPackageFactory:

    def __init__(self):
        pass

    @abc.abstractmethod
    def create_data_package(self) -> SampleDataPackage:
        pass

    @abc.abstractmethod
    def create_data_package_iter(self, batch_size: int = 100) -> Generator[SampleDataPackage, None, None]:
        pass



import abc
import logging
from typing import Dict, List, Any, Optional

import pandas as pd

from genomics_data_index.api.query.features.FeaturesComparator import FeaturesComparator
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.QueryFeature import QueryFeature
from genomics_data_index.storage.model.QueryFeatureFactory import QueryFeatureFactory
from genomics_data_index.storage.model.db import FeatureSamples
from genomics_data_index.storage.service.SampleService import SampleService

logger = logging.getLogger(__name__)


class FeatureSamplesSummarizer(abc.ABC):

    def __init__(self):
        pass

    @abc.abstractmethod
    def summary_data(self, samples: SampleSet, unknown_samples: Optional[SampleSet], total: int) -> List[Any]:
        pass

    @abc.abstractmethod
    def summary_names(self) -> List[str]:
        pass


class FeatureSamplesSingleCategorySummarizer(FeatureSamplesSummarizer):
    SUMMARY_NAMES = ['Count', 'Unknown Count', 'Present and Unknown Count',
                     'Total', 'Percent', 'Unknown Percent', 'Present and Unknown Percent']

    def __init__(self):
        super().__init__()

    def summary_data(self, samples: SampleSet, unknown_samples: Optional[SampleSet], total: int) -> List[Any]:
        sample_count = len(samples)

        if unknown_samples is not None:
            unknown_sample_count = len(unknown_samples)
            present_and_unknown_sample_count = len(samples.union(unknown_samples))
            unknown_percent = (unknown_sample_count / total) * 100
            present_and_unknown_percent = (present_and_unknown_sample_count / total) * 100
        else:
            unknown_sample_count = pd.NA
            present_and_unknown_sample_count = pd.NA
            unknown_percent = pd.NA
            present_and_unknown_percent = pd.NA

        percent = (sample_count / total) * 100
        return [sample_count, unknown_sample_count, present_and_unknown_sample_count, total,
                percent, unknown_percent, present_and_unknown_percent]

    def summary_names(self) -> List[str]:
        return self.SUMMARY_NAMES


class FeatureSamplesMultipleCategorySummarizer(FeatureSamplesSummarizer):
    COMPARE_TYPE_LABELS = ['_', '_Unknown ', '_Present and Unknown ']

    def __init__(self, sample_categories: List[SampleSet], category_prefixes: List[str] = None,
                 compare_kind: str = 'percent'):
        super().__init__()

        self._use_count = False
        self._factor = None
        if compare_kind not in ['proportion', 'percent', 'count']:
            raise Exception(f'compare_kind={compare_kind} must be one of "percent", "proportion", or "count"')
        elif compare_kind == 'proportion':
            self._factor = 1
        elif compare_kind == 'percent':
            self._factor = 100
        else:
            self._use_count = True

        if category_prefixes is None:
            category_prefixes = [f'Category{x + 1}' for x in range(len(sample_categories))]
        elif not isinstance(category_prefixes, list):
            raise Exception(f'category_names={category_prefixes} must be a list or None')
        elif len(category_prefixes) != len(sample_categories):
            raise Exception(f'sample_categories has {len(sample_categories)} elements but, '
                            f'category_names has {len(category_prefixes)} elements. These must be the same size.')
        else:
            # I convert to string in case someone passes a list that's not composed of strings
            category_prefixes = [str(c) for c in category_prefixes]

        self._category_prefixes = category_prefixes
        self._prefixes_sample_categories = {p: c for p, c in zip(category_prefixes, sample_categories)}

        self._summary_names = ['Total'] \
                              + [f'{c}{t}{compare_kind}' for t in self.COMPARE_TYPE_LABELS for c in
                                 self._category_prefixes] \
                              + [f'{c}_total' for c in self._category_prefixes]

    def summary_data(self, samples: SampleSet, unknown_samples: Optional[SampleSet], total: int) -> List[Any]:
        data = [total]
        present_data = []
        unknown_data = []
        present_and_unknown_data = []
        for category_prefix in self._category_prefixes:
            sample_category = self._prefixes_sample_categories[category_prefix]
            sample_category_total = len(sample_category)

            samples_in_category = samples.intersection(sample_category)
            category_count = len(samples_in_category)

            if unknown_samples is not None:
                unknown_samples_in_category = unknown_samples.intersection(sample_category)
                unknown_sample_count = len(unknown_samples_in_category)
                present_and_unknown_sample_count = len(samples_in_category.union(unknown_samples_in_category))
            else:
                unknown_sample_count = pd.NA
                present_and_unknown_sample_count = pd.NA

            if self._use_count:
                present_data.append(category_count)
                unknown_data.append(unknown_sample_count)
                present_and_unknown_data.append(present_and_unknown_sample_count)
            else:
                if sample_category_total > 0:
                    category_percent = (category_count / sample_category_total) * self._factor

                    if unknown_samples is not None:
                        unknown_sample_percent = (unknown_sample_count / sample_category_total) * self._factor
                        present_and_unknown_sample_percent = (present_and_unknown_sample_count / sample_category_total) \
                                                             * self._factor
                    else:
                        unknown_sample_percent = pd.NA
                        present_and_unknown_sample_percent = pd.NA
                else:
                    category_percent = pd.NA
                    unknown_sample_percent = pd.NA
                    present_and_unknown_sample_percent = pd.NA

                present_data.append(category_percent)
                unknown_data.append(unknown_sample_percent)
                present_and_unknown_data.append(present_and_unknown_sample_percent)

        # Add data in proper order
        for category_data in [present_data, unknown_data, present_and_unknown_data]:
            data.extend(category_data)

        # Append totals for each category to end
        for category_prefix in self._category_prefixes:
            sample_category = self._prefixes_sample_categories[category_prefix]
            sample_category_total = len(sample_category)
            data.append(sample_category_total)

        return data

    def summary_names(self) -> List[str]:
        return self._summary_names


class FeaturesFromIndexComparator(FeaturesComparator, abc.ABC):

    def __init__(self, connection: DataIndexConnection, include_unknown_samples: bool,
                 include_unknown_no_present_samples: bool):
        super().__init__(connection=connection, include_unknown_samples=include_unknown_samples,
                         include_unknown_no_present_samples=include_unknown_no_present_samples)

    def _get_samples_or_empty(self, feature: QueryFeature,
                              feature_id_set_dict: Dict[str, SampleSet]) -> SampleSet:
        if feature.id in feature_id_set_dict:
            return feature_id_set_dict[feature.id]
        else:
            return SampleSet.create_empty()

    def _create_summary_df(self, present_features: Dict[str, FeatureSamples],
                           present_samples: SampleSet,
                           feature_samples_summarizer: FeatureSamplesSummarizer) -> pd.DataFrame:
        data = []
        total = self._get_total(present_samples)
        for feature_id in present_features:
            feature = present_features[feature_id]

            samples_in_feature = present_samples.intersection(feature.sample_ids)
            sample_count = len(samples_in_feature)

            # We don't want to include results with 0 for sample count
            # (unless we need to check for unknown sample counts)
            if sample_count > 0 or self._include_unknown_no_present_samples:
                if self._include_unknown_samples and not feature.is_unknown:
                    query_feature = QueryFeatureFactory.instance().create_feature(feature.query_id)
                    unknown_samples_dict = self._connection.sample_service.find_unknown_sample_sets_by_features(
                        [query_feature])
                    samples_unknown_in_feature = present_samples.intersection(
                        self._get_samples_or_empty(query_feature, unknown_samples_dict))
                    sample_unknown_count = len(samples_unknown_in_feature)

                    extra_msg = f' for feature {query_feature}'
                    samples_in_feature = self.remove_unknowns_from_present_samples(
                        sample_set=samples_in_feature,
                        unknown_set=samples_unknown_in_feature,
                        sample_service=self._connection.sample_service,
                        extra_message=extra_msg,
                    )
                else:
                    samples_unknown_in_feature = None
                    sample_unknown_count = None

                sample_count = len(samples_in_feature)

                # Since "sample_count" may have been adjusted due to unknown samples, need to check
                # for 0 sample_count again
                if sample_count > 0 or (self._include_unknown_no_present_samples and
                                        sample_unknown_count is not None and sample_unknown_count > 0):
                    data.append(self._create_feature_sample_count_row(feature_id,
                                                                      feature=feature,
                                                                      feature_samples=samples_in_feature,
                                                                      feature_samples_unknown=samples_unknown_in_feature,
                                                                      total=total,
                                                                      feature_samples_summarizer=feature_samples_summarizer)
                                )
        summary_df = pd.DataFrame(data,
                                  columns=[self.index_name] \
                                          + self.feature_id_columns + feature_samples_summarizer.summary_names())
        return summary_df.set_index(self.index_name)

    def summary(self, sample_set: SampleSet) -> pd.DataFrame:
        samples_summarizer = FeatureSamplesSingleCategorySummarizer()
        return self._do_summary(sample_set=sample_set, feature_samples_summarizer=samples_summarizer)

    def features_comparison(self, selected_samples: SampleSet,
                            sample_categories: List[SampleSet],
                            category_prefixes: List[str] = None,
                            category_samples_threshold: int = None,
                            use_only_samples_in_categories: bool = True,
                            unit: str = 'percent') -> pd.DataFrame:
        sample_categories_in_selected = [c.intersection(selected_samples) for c in sample_categories]

        if category_samples_threshold is not None:
            filtered_categories = []
            filtered_category_prefixes = []
            for sample_category, prefix in zip(sample_categories_in_selected, category_prefixes):
                if len(sample_category) >= category_samples_threshold:
                    filtered_categories.append(sample_category)
                    filtered_category_prefixes.append(prefix)
        else:
            filtered_categories = sample_categories_in_selected
            filtered_category_prefixes = category_prefixes

        filtered_samples_union = SampleSet.create_empty()

        if use_only_samples_in_categories:
            for sample_category in filtered_categories:
                filtered_samples_union = filtered_samples_union.union(sample_category)
            samples_for_summary = filtered_samples_union
        else:
            samples_for_summary = selected_samples

        samples_summarizer = FeatureSamplesMultipleCategorySummarizer(sample_categories=filtered_categories,
                                                                      category_prefixes=filtered_category_prefixes,
                                                                      compare_kind=unit)
        return self._do_summary(sample_set=samples_for_summary, feature_samples_summarizer=samples_summarizer)

    @abc.abstractmethod
    def _do_summary(self, sample_set: SampleSet, feature_samples_summarizer: FeatureSamplesSummarizer) -> pd.DataFrame:
        pass

    @abc.abstractmethod
    def _create_feature_sample_count_row(self, feature_id: str, feature: FeatureSamples,
                                         feature_samples: SampleSet,
                                         feature_samples_unknown: Optional[SampleSet],
                                         total: int,
                                         feature_samples_summarizer: FeatureSamplesSummarizer) -> List[Any]:
        pass

    @classmethod
    def remove_unknowns_from_present_samples(cls, sample_set: SampleSet, unknown_set: SampleSet,
                                             sample_service: SampleService, extra_message: str = '') -> SampleSet:
        # Handle situations where a sample is both found and unknown
        # This should not occur (unless there's bugs) but this will warn you if it does
        common_unknown_found_set = sample_set.intersection(unknown_set)
        if len(common_unknown_found_set) > 0:
            # Some of the code here is to only print the top "max" number of sample names in the warning message
            max = 10
            common_ids = list(common_unknown_found_set)[:max]
            common_names = [s.name for s in sample_service.find_samples_by_ids(common_ids)]
            if len(common_unknown_found_set) > max:
                msg = f'names=[{", ".join(common_names)}, ...]'
            else:
                msg = f'names={common_names}'
            logger.warning(f'There are {len(common_unknown_found_set)} samples ({msg}) that are both unknown and found'
                           f'{extra_message}. Will set these samples to unknown.')
            sample_set = sample_set.minus(common_unknown_found_set)

        return sample_set



from __future__ import annotations

import abc
import logging
import uuid
from pathlib import Path

from genomics_data_index.storage.util import TRACE_LEVEL

logger = logging.getLogger(__name__)

persistence_namespace = uuid.UUID('260bb0a7-6573-4282-bcdb-f4008aa42af8')


def generate_sample_persistence_name(sample_name: str) -> str:
    return uuid.uuid3(namespace=persistence_namespace, name=sample_name).hex


class SampleData(abc.ABC):

    def __init__(self, sample_name: str):
        self._name = sample_name
        self._sample_name_persistence = generate_sample_persistence_name(sample_name)

    @property
    def sample_name(self) -> str:
        return self._name

    @property
    def sample_name_persistence(self) -> str:
        return self._sample_name_persistence

    @abc.abstractmethod
    def is_preprocessed(self) -> bool:
        pass

    def preprocess(self, output_dir: Path) -> SampleData:
        return self.persist(output_dir)

    def persist(self, output_dir: Path) -> SampleData:
        if not self.is_preprocessed():
            logger.log(TRACE_LEVEL, f'Processing sample [{self.sample_name}] and saving to [{output_dir}]')
            return self._do_preprocess_and_persist(output_dir)
        else:
            return self._do_persist(output_dir)

    @abc.abstractmethod
    def _do_preprocess_and_persist(self, output_dir: Path) -> SampleData:
        pass

    @abc.abstractmethod
    def _do_persist(self, output_dir: Path) -> SampleData:
        pass



import pytest

from genomics_data_index.storage.io.mlst.MLSTSistrReader import MLSTSistrReader
from genomics_data_index.test.integration import sistr_mlst_file, expand_list_by


@pytest.fixture
def mlst_reader() -> MLSTSistrReader:
    return MLSTSistrReader(mlst_file=sistr_mlst_file)


def test_get_features_table(mlst_reader):
    num_samples = 2
    num_loci = 330

    mlst_df = mlst_reader.get_features_table()

    assert ['File', 'Sample', 'Scheme', 'Locus', 'Allele'] == list(mlst_df.columns)

    assert num_samples * num_loci == len(mlst_df)
    assert expand_list_by(['GCF_000006945.2_ASM694v2_genomic', 'GCF_000007545.1_ASM754v1_genomic'],
                          num_loci) == list(mlst_df['File'].tolist())
    assert expand_list_by(['GCF_000006945', 'GCF_000007545'],
                          num_loci) == list(mlst_df['Sample'].tolist())

    assert {'sistr_330'} == set(mlst_df['Scheme'].tolist())

    assert ['3371366009'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000006945') &
                                              (mlst_df['Locus'] == 'NZ_AOXE01000059.1_60'), 'Allele'].tolist())

    assert ['3848099890'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000007545') &
                                              (mlst_df['Locus'] == 'NZ_AOXE01000019.1_14'), 'Allele'].tolist())


def test_get_features_table_missing_feature(mlst_reader):
    mlst_df = mlst_reader.get_features_table()

    assert ['File', 'Sample', 'Scheme', 'Locus', 'Allele'] == list(mlst_df.columns)
    assert {'sistr_330'} == set(mlst_df['Scheme'].tolist())

    assert ['?'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000006945') &
                                     (mlst_df['Locus'] == 'NZ_AOXE01000085.1_58'), 'Allele'].tolist())

    assert ['3038970538'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000007545') &
                                              (mlst_df['Locus'] == 'NZ_AOXE01000085.1_58'), 'Allele'].tolist())


def test_get_features_table_invalid_feature(mlst_reader):
    mlst_df = mlst_reader.get_features_table()

    assert ['File', 'Sample', 'Scheme', 'Locus', 'Allele'] == list(mlst_df.columns)
    assert {'sistr_330'} == set(mlst_df['Scheme'].tolist())

    assert ['?'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000006945') &
                                     (mlst_df['Locus'] == 'NZ_AOXE01000034.1_134'), 'Allele'].tolist())

    assert ['682273023'] == list(mlst_df.loc[(mlst_df['Sample'] == 'GCF_000007545') &
                                             (mlst_df['Locus'] == 'NZ_AOXE01000034.1_134'), 'Allele'].tolist())



from __future__ import annotations

import logging
from pathlib import Path
from typing import Optional, Dict

import yaml

logger = logging.getLogger(__name__)


class ConfigFile:

    def __init__(self, config_dict: Optional[Dict[str, str]] = None):
        if config_dict is None:
            config_dict = {}
        self._config = config_dict

    @property
    def database_connection(self) -> str:
        return self._config['database_connection']

    @database_connection.setter
    def database_connection(self, connection: str) -> None:
        self._config['database_connection'] = connection

    @property
    def database_dir(self) -> Path:
        return Path(self._config['database_dir'])

    @database_dir.setter
    def database_dir(self, dir: Path) -> None:
        self._config['database_dir'] = str(dir)

    @property
    def sqlite_database(self) -> Path:
        return Path(self._config['sqlite_database'])

    @sqlite_database.setter
    def sqlite_database(self, db_path: Path):
        self._config['sqlite_database'] = str(db_path)

    def write(self, file: Path) -> None:
        """
        Writes The configuration to the passed file.
        :param file: The file where configuration should be written.
        :return: None.
        """
        if file.exists():
            raise Exception(f'File [{file}] already exists')

        with open(file, 'w') as fh:
            yaml.dump(self._config, fh)

        logger.debug(f'Wrote configuration to [{file}]')

    @classmethod
    def read_config(self, file: Path) -> ConfigFile:
        if not file.exists():
            raise Exception(f'Config file {file} does not exist')

        logger.debug(f'Reading configuration from {file}')

        with open(file) as file:
            config = yaml.load(file, Loader=yaml.FullLoader)
            if config is None:
                config = {}

            if 'database_connection' not in config and 'sqlite_database' not in config:
                logger.warning(f'Missing database_connection and sqlite_database in config file {file}')

            if 'database_connection' not in config:
                config['database_connection'] = None
            if 'sqlite_database' not in config:
                config['sqlite_database'] = None
            if 'database_dir' not in config:
                logger.warning(f'Missing database_dir in config file {file}')
                config['database_dir'] = None

            return ConfigFile(config_dict=config)



import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Any

from genomics_data_index.configuration.connector.FilesystemStorage import FilesystemStorage
from genomics_data_index.storage.io.mutation.NucleotideSampleDataPackage import NucleotideSampleDataPackage
from genomics_data_index.storage.model.db import DatabasePathTranslator, SampleNucleotideVariation, Sample, \
    SampleKmerIndex
from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service.KmerService import KmerService
from genomics_data_index.storage.service.ReferenceService import ReferenceService
from genomics_data_index.storage.service.SampleService import SampleService
from genomics_data_index.storage.service.VariationService import VariationService
from genomics_data_index.test.integration import reference_file
from genomics_data_index.test.integration import sourmash_signatures


def setup_services(root_dir: Path, database_file: Path) -> Dict[str, Any]:
    filesystem_storage = FilesystemStorage(root_dir)
    dpt = DatabasePathTranslator(filesystem_storage.root_dir)
    db_connection = DatabaseConnection(f'sqlite:///{database_file}', dpt)
    reference_service = ReferenceService(db_connection, filesystem_storage.reference_dir)
    sample_service = SampleService(db_connection, sql_select_limit=500)
    variation_service = VariationService(database_connection=db_connection,
                                         reference_service=reference_service,
                                         sample_service=sample_service,
                                         variation_dir=filesystem_storage.variation_dir,
                                         index_unknown_missing=False,
                                         sql_select_limit=50)
    kmer_service = KmerService(database_connection=db_connection,
                               sample_service=sample_service,
                               features_dir=filesystem_storage.kmer_dir)

    return {
        'reference_service': reference_service,
        'variation_service': variation_service,
        'database_connection': db_connection,
        'kmer_service': kmer_service
    }


def test_path_translation_nucleotide_data(snippy_nucleotide_data_package: NucleotideSampleDataPackage):
    with TemporaryDirectory() as root_dir_1_str:
        root_dir_1 = Path(root_dir_1_str)
        data_dir_1 = root_dir_1 / 'data'
        database_file = root_dir_1 / 'db.sqlite'
        services = setup_services(data_dir_1, database_file)
        variation_service = services['variation_service']
        reference_service = services['reference_service']
        database_connection = services['database_connection']

        # Insert data
        reference_service.add_reference_genome(reference_file)
        variation_service.insert(feature_scope_name='genome',
                                 data_package=snippy_nucleotide_data_package)

        # Get a single saved object which has attached files on the filesystem
        sampleA_nv = database_connection.get_session().query(SampleNucleotideVariation) \
            .join(SampleNucleotideVariation.sample) \
            .filter(Sample.name == 'SampleA').one()

        # Make sure this file exists and is in the correct location
        nvf: Path = sampleA_nv.nucleotide_variants_file
        bf: Path = sampleA_nv.masked_regions_file
        assert nvf.exists()
        assert nvf.parent.parent == data_dir_1
        assert bf.exists()
        assert bf.parent.parent == data_dir_1

        # Now we move data files into brand new directory
        with TemporaryDirectory() as root_dir_2_str:
            root_dir_2 = Path(root_dir_2_str)
            data_dir_2 = root_dir_2 / 'data'
            shutil.move(data_dir_1, data_dir_2)

            assert not nvf.exists(), 'Data should be moved so original file not exist'
            assert not bf.exists(), 'Data should be moved so original file not exist'

            services_2 = setup_services(data_dir_2, database_file)
            database_connection_2 = services_2['database_connection']

            # Get a single saved object which has attached files on the filesystem
            sampleA_nv_2 = database_connection_2.get_session().query(SampleNucleotideVariation) \
                .join(SampleNucleotideVariation.sample) \
                .filter(Sample.name == 'SampleA').one()
            nvf_2: Path = sampleA_nv_2.nucleotide_variants_file
            bf_2: Path = sampleA_nv_2.masked_regions_file

            assert nvf_2.exists(), 'Path from database should now correspond to moved path'
            assert nvf_2.parent.parent == data_dir_2
            assert bf_2.exists()
            assert bf_2.parent.parent == data_dir_2


def test_path_translation_kmer_data(snippy_nucleotide_data_package: NucleotideSampleDataPackage):
    with TemporaryDirectory() as root_dir_1_str:
        root_dir_1 = Path(root_dir_1_str)
        data_dir_1 = root_dir_1 / 'data'
        database_file = root_dir_1 / 'db.sqlite'

        services = setup_services(data_dir_1, database_file)
        kmer_service = services['kmer_service']
        database_connection = services['database_connection']

        # Insert data
        for sample_name in sourmash_signatures:
            kmer_service.insert_kmer_index(sample_name=sample_name,
                                           kmer_index_path=sourmash_signatures[sample_name])

        # Get a single saved object which has attached files on the filesystem
        sampleA_kmer = database_connection.get_session().query(SampleKmerIndex) \
            .join(SampleKmerIndex.sample) \
            .filter(Sample.name == 'SampleA').one()

        # Make sure this file exists and is in the correct location
        kf: Path = sampleA_kmer.kmer_index_path
        assert kf.exists()
        assert kf.parent.parent == data_dir_1

        # Now we move data files into brand new directory
        with TemporaryDirectory() as root_dir_2_str:
            root_dir_2 = Path(root_dir_2_str)
            data_dir_2 = root_dir_2 / 'data'
            shutil.move(data_dir_1, data_dir_2)

            assert not kf.exists(), 'Data should be moved so original file not exist'

            services_2 = setup_services(data_dir_2, database_file)
            database_connection_2 = services_2['database_connection']

            # Get a single saved object which has attached files on the filesystem
            sampleA_kmer_2 = database_connection_2.get_session().query(SampleKmerIndex) \
                .join(SampleKmerIndex.sample) \
                .filter(Sample.name == 'SampleA').one()
            kf_2: Path = sampleA_kmer_2.kmer_index_path

            assert kf_2.exists(), 'Path from database should now correspond to moved path'
            assert kf_2.parent.parent == data_dir_2



import tempfile
from pathlib import Path

from sourmash import load_file_as_signatures

from genomics_data_index.storage.index.KmerIndexer import KmerIndexerSourmash
from genomics_data_index.test.integration import data_dir
from genomics_data_index.test.integration.storage.index import get_values_from_signatures


def test_index_single_file_uncompressed():
    sampleA = data_dir / 'SampleA' / 'snps.aligned.nogap.fa.gz'

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        index_out = tmp_path / 'SampleA.sig'

        kmer_indexer = KmerIndexerSourmash(k=31, scaled=1000, compress=False)
        indexed_path = kmer_indexer.index('SampleA', index_out, [sampleA])

        assert indexed_path == index_out
        assert indexed_path.exists()
        sigs = load_file_as_signatures(str(indexed_path))
        values = get_values_from_signatures(sigs)

        assert {31} == values['ksize']
        assert {1000} == values['scaled']


def test_index_single_file_compressed():
    sampleA = data_dir / 'SampleA' / 'snps.aligned.nogap.fa.gz'

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        index_out = tmp_path / 'SampleA.sig'

        kmer_indexer = KmerIndexerSourmash(k=31, scaled=1000, compress=True)
        indexed_path = kmer_indexer.index('SampleA', index_out, [sampleA])

        assert '.gz' == indexed_path.suffix
        assert indexed_path.exists()
        sigs = load_file_as_signatures(str(indexed_path))
        values = get_values_from_signatures(sigs)

        assert {31} == values['ksize']
        assert {1000} == values['scaled']


def test_index_single_file_multiple_k():
    sampleA = data_dir / 'SampleA' / 'snps.aligned.nogap.fa.gz'

    with tempfile.TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir)
        index_out = tmp_path / 'SampleA.sig.gz'

        kmer_indexer = KmerIndexerSourmash(k=[21, 31], scaled=1000, compress=True)
        indexed_path = kmer_indexer.index('SampleA', index_out, [sampleA])

        assert indexed_path.exists()
        sigs = load_file_as_signatures(str(indexed_path))
        values = get_values_from_signatures(sigs)

        assert {21, 31} == values['ksize']
        assert {1000} == values['scaled']



from __future__ import annotations

import abc
import logging
from typing import Union, List, cast, Callable

import pandas as pd
from ete3 import Tree, TreeStyle, NodeStyle

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.WrappedSamplesQuery import WrappedSamplesQuery
from genomics_data_index.api.viewer.TreeStyler import TreeStyler, HighlightStyle
from genomics_data_index.configuration.connector import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet

logger = logging.getLogger(__name__)


class TreeSamplesQuery(WrappedSamplesQuery, abc.ABC):
    """
    A class to handle queries that are joined to a tree. Subclasses are used for different implementations
    depending on what sort of data was used to build the tree (e.g., mutations or kmers).
    """

    BUILD_TREE_KINDS = ['mutation']
    ISIN_TREE_TYPES = ['distance', 'mrca']

    def __init__(self, connection: DataIndexConnection, wrapped_query: SamplesQuery, tree: Tree):
        """
        Builds a new TreeSamplesQuery from the given information. In most normal operations SamplesQuery objects
        are not created directly but are instead created from an :py:class:`genomics_data_index.api.GenomicsDataIndex`
        object or from operations applied to a SamplesQuery (e.g., build_tree() or join_tree()).

        :param connection: A connection to a database containing samples.
        :param wrapped_query: The SamplesQuery to wrap around and join to the passed tree.
        :param tree: The tree to join to this query.
        :return: A new TreeSamplesQuery object.
        """
        super().__init__(connection=connection, wrapped_query=wrapped_query)
        self._tree = tree

    def _within_mrca(self, data: Union[str, List[str], SamplesQuery, SampleSet]) -> SamplesQuery:
        if isinstance(data, str):
            data = [data]

        sample_names, query_infix = self._get_sample_names_query_infix_from_data(data)
        if len(sample_names) == 0:
            found_sample_ids = SampleSet.create_empty()
        else:
            sample_leaves_list = []
            for name in sample_names:
                sample_leaves = self._tree.get_leaves_by_name(name)
                if len(sample_leaves) != 1:
                    raise Exception(
                        f'Invalid number of matching leaves for sample [{name}], leaves {sample_leaves}')
                else:
                    sample_leaves_list.append(sample_leaves[0])

            if len(sample_leaves_list) == 0:
                raise Exception(f'Should at least have some leaves in the tree matching data={data}')
            elif len(sample_leaves_list) == 1:
                found_sample_names = [sample_leaves_list[0].name]
            else:
                # According to https://github.com/etetoolkit/ete/issues/503 'get_common_ancestor' does *not*
                # Include the first node calling the function so I have to leave it in the list.
                first_sample_leaf = sample_leaves_list[0]
                ancestor_node = first_sample_leaf.get_common_ancestor(sample_leaves_list)
                found_sample_names = ancestor_node.get_leaf_names()

            # Ignore not found since the reference genome in the tree will be unlikely to be recorded as a sample
            # in the database
            found_sample_ids = self._query_connection.sample_service.get_sample_set_by_names(found_sample_names,
                                                                                             ignore_not_found=True)
        return self.intersect(found_sample_ids, f'within(mrca of {query_infix})')

    def _isin_internal(self, data: Union[str, List[str], SamplesQuery, SampleSet], kind: str, **kwargs) -> SamplesQuery:
        if kind == 'distance':
            return self._within_distance(data=data, **kwargs)
        elif kind == 'mrca':
            return self._within_mrca(data=data)
        else:
            raise Exception(f'kind=[{kind}] is not supported. Must be one of {self._isin_kinds()}')

    def build_tree(self, kind: str, **kwargs) -> SamplesQuery:
        samples_query = self._wrapped_query.build_tree(kind=kind, **kwargs)

        if isinstance(samples_query, TreeSamplesQuery):
            samples_query = cast(TreeSamplesQuery, samples_query)
            return samples_query._wrap_create(self)
        else:
            raise Exception(f'Build tree is not of type {TreeSamplesQuery.__class__}')

    def join_tree(self, tree: Tree, kind='mutation', **kwargs) -> SamplesQuery:
        samples_query = self._wrapped_query.join_tree(tree=tree, kind=kind, **kwargs)

        if isinstance(samples_query, TreeSamplesQuery):
            samples_query = cast(TreeSamplesQuery, samples_query)
            return samples_query._wrap_create(self)
        else:
            raise Exception(f'Build tree is not of type {TreeSamplesQuery.__class__}')

    def _tree_copy(self) -> Tree:
        return self._tree.copy(method='newick')

    def set_outgroup(self, sample_name: str) -> SamplesQuery:
        """
        Sets the outgroup of the tree to the specified sample name.
        :param sample_name: The name to set as an outgroup. Must exist as a leaf in the tree.
        :return: A new TreeSamplesQuery with the defined outgroup set.
        """
        tree = self._tree_copy()
        tree.set_outgroup(sample_name)
        return self._create_from_tree_internal(tree)

    def prune(self, preserve_branch_length: bool = True, include_present: bool = True,
              include_unknown: bool = False, include_absent: bool = False) -> SamplesQuery:
        """
        Prunes the tree joined to this TreeSamplesQuery down to only the currently selected samples.
        :param preserve_branch_length: True if branch lengths between the subset of samples should be preserved, False otherwise.
        :param include_present: If samples with a present status should be included in the tree.
        :param include_unknown: If samples with an unknown status should be included in the tree.
        :param include_absent: If samples with an absent status from query should be included in the tree.
        :return: A new TreeSamplesQuery with the tree pruned down to only the selected samples (and the reference
                 genome if it is one of the leaves of the tree).
        """
        tree = self._tree_copy_prune(preserve_branch_length=preserve_branch_length,
                                     include_present=include_present, include_unknown=include_unknown,
                                     include_absent=include_absent)
        return self._create_from_tree_internal(tree)

    def _tree_copy_prune(self,
                         preserve_branch_length: bool,
                         include_present: bool,
                         include_unknown: bool,
                         include_absent: bool,
                         from_query: SamplesQuery = None,
                         ) -> Tree:
        tree = self._tree_copy()
        if from_query is not None:
            query = from_query
        else:
            query = self

        nodes_to_keep = self._get_samples_to_keep_from_query(query, include_present=include_present,
                                                             include_unknown=include_unknown,
                                                             include_absent=include_absent)
        tree.prune(nodes_to_keep, preserve_branch_length=preserve_branch_length)
        return tree

    def _get_samples_to_keep_from_query(self, query: SamplesQuery,
                                        include_present: bool,
                                        include_unknown: bool,
                                        include_absent: bool,
                                        ) -> List[str]:
        return query.tolist(names=True, include_present=include_present, include_unknown=include_unknown,
                            include_absent=include_absent)

    @abc.abstractmethod
    def _create_from_tree_internal(self, tree: Tree) -> SamplesQuery:
        pass

    def _within_distance(self, data: Union[str, List[str], SamplesQuery, SampleSet], distance: float,
                         units: str, **kwargs) -> SamplesQuery:
        if self._can_handle_distance_units(units):
            return self._within_distance_internal(data=data,
                                                  distance=distance, units=units)
        else:
            return self._wrap_create(self._wrapped_query._within_distance(data=data,
                                                                          distance=distance,
                                                                          units=units,
                                                                          **kwargs))

    @abc.abstractmethod
    def _within_distance_internal(self, data: Union[str, List[str], SamplesQuery, SampleSet], distance: float,
                                  units: str) -> SamplesQuery:
        pass

    @abc.abstractmethod
    def _can_handle_distance_units(self, units: str) -> bool:
        pass

    def _isin_kinds(self) -> List[str]:
        return list(set(super()._isin_kinds() + self.ISIN_TREE_TYPES))

    def _can_handle_isin_kind(self, kind: str) -> bool:
        return kind in self.ISIN_TREE_TYPES

    def tree_styler(self,
                    initial_style: TreeStyle = None,
                    mode='r',
                    highlight_style: Union[str, HighlightStyle] = 'pastel',
                    node_style: NodeStyle = None,
                    legend_nsize: int = 20, legend_fsize: int = 11,
                    annotate_color_present: str = 'black',
                    annotate_color_absent: str = 'white',
                    annotate_color_unknown: str = 'lightgray',
                    annotate_opacity_present: float = 1.0,
                    annotate_opacity_absent: float = 0.0,
                    annotate_opacity_unknown: float = 1.0,
                    annotate_border_color: str = 'black',
                    annotate_kind: str = 'rect',
                    annotate_box_width: int = 30,
                    annotate_box_height: int = 30,
                    annotate_border_width: int = 1,
                    annotate_margin: int = 0,
                    annotate_guiding_lines: bool = True,
                    annotate_guiding_lines_color: str = 'gray',
                    figure_margin: int = None,
                    show_border: bool = True,
                    title: str = None,
                    title_fsize: int = 16,
                    legend_title: str = None,
                    annotate_show_box_label: bool = False,
                    annotate_box_label_color: str = 'white',
                    annotate_arc_span: int = 350,
                    annotate_label_fontsize: int = 12,
                    include_unknown: bool = True,
                    show_leaf_names: bool = True,
                    leaf_name_fontsize: int = 12,
                    leaf_name_func: Callable[[str, pd.Series], str] = None,
                    show_legend_type_labels: bool = True,
                    legend_type_label_present: str = 'Pr.',
                    legend_type_label_unknown: str = 'Un.',
                    rotation: float = 0,
                    allow_face_overlap: bool = False,
                    show_branch_length: bool = False,
                    show_branch_support: bool = False,
                    tree_line_width: int = None,
                    tree_scale: float = None,
                    prune_tree: bool = True) -> TreeStyler:
        """
        Constructs a new :py:class:`genomics_data_index.api.viewer.TreeStyler` object used to style and visualize trees.
        All parameters listed below are optional.
        :param initial_style: The initial ete3.TreeStyle to start with.
        :param mode: Either 'r' (rectangular) or 'c' (circular).
        :param highlight_style: A style used to define how the highlight() method should behave.
                                Can either be one of the named highlight styles ['light', 'light_hn', 'pastel', 'medium', dark']
                                or an instance of a :py:class:`genomics_data_index.api.viewer.TreeStyler.HighlightStyle`.
        :param node_style: The default ete3.NodeStyle object for styling the nodes of the tree.
        :param legend_nsize: The legend node size.
        :param legend_fsize: The legend font size.
        :param annotate_color_present: The default color of samples which are present in the set for the annotate() method.
        :param annotate_color_absent: The default color of samples which are absent in the set for the annotate() method.
        :param annotate_color_unknown: The default color of samples which are unknown if present/absent in the set for the annotate() method.
        :param annotate_opacity_present: The default opacity of samples which are present in the set for the annotate() method.
        :param annotate_opacity_absent: The default opacity of samples which are absent in the set for the annotate() method.
        :param annotate_opacity_unknown: The default opacity of samples which are unknown if present/absent in the set for the annotate() method.
        :param annotate_border_color: The default border color of the drawn annotations.
        :param annotate_kind: The default kind color of the drawn annotations (either 'circle' or 'rectangle').
        :param annotate_box_width: The width of the boxes for the drawn annotations.
        :param annotate_box_height: The height of the boxes for the drawn annotations.
        :param annotate_border_width: The width of the border for the boxes for the drawn annotations.
        :param annotate_margin: The margin width of the boxes for the drawn annotations.
        :param annotate_guiding_lines: True if guiding lines should be drawn that matches sample names to the annotation boxes,
                                       False otherwise.
        :param annotate_guiding_lines_color:  The color of the annotate guiding lines.
        :param figure_margin: The margin spacing (used for all of top, bottom, left, and right) for the overall figure.
        :param show_border: True if a border should be shown around the overall figure, False otherwise.
        :param title: A title for the figure.
        :param title_fsize: The font size of the figure title.
        :param legend_title: The title of the legend.
        :param annotate_show_box_label: True if labels should be shown in the annotation boxes, False otherwise.
        :param annotate_box_label_color: The color of the labels in the annotation boxes.
        :param annotate_arc_span: For mode='c' (circular) the degrees the circular tree should span.
        :param annotate_label_fontsize: The font size of the annotation labels.
        :param include_unknown: Whether or not to include unknowns in annotations/highlights.
        :param show_leaf_names: True if leaf names should be shown on the tree, False otherwise.
        :param leaf_name_fontsize: The font size of leaf names.
        :param leaf_name_func: A function which lets you create custom leaf names to display.
                               The function should look like: func(name: str, metadata: pd.Series) -> str.
                               That is it takes as input the leaf name and a pandas.Series of metadata for the
                               particular sample (derived from the query.toframe() set of data).
                               For example: tree_styler(...,
                                 leaf_name_func=lambda name, metadata: f'{name}_{metadata["Location"]}')
                               This would display a label like "SampleX_Canada" for each sample (assumes that "Location"
                               is a column name in the table produced by query.toframe()).
        :param show_legend_type_labels: Whether or not to show labels for legend types/categories (present or unknown).
        :param legend_type_label_present: Text to show above legend color for present items.
        :param legend_type_label_unknown: Text to show above legend color for unknown items.
        :param rotation: The rotation of the tree in degrees (for circular mode).
        :param allow_face_overlap: Allow overlap in node faces for circular images.
        :param show_branch_length: Show branch lengths.
        :param show_branch_support: Show branch supports.
        :param tree_line_width: The line width for the tree. Overrides 'hz_line_width' and 'vt_line_width'
                                in node_style. Default: None (no overriding of line width in node_style).
        :param tree_scale: A scale factor for the tree.
        :param prune_tree: Whether or not to prune the tree in this query before creating a
                           TreeStyler object. Default: True.
        :return: A new :py:class:`genomics_data_index.api.viewer.TreeStyler` object used to style and visualize trees.
        """
        if show_leaf_names and leaf_name_func is not None:
            sample_metadata = self.toframe(include_present=True, include_unknown=True)
            sample_metadata = sample_metadata.set_index('Sample Name')
        else:
            sample_metadata = None

        if prune_tree:
            tree = self._tree_copy_prune(preserve_branch_length=True,
                                         include_present=True, include_unknown=True,
                                         include_absent=False)
        else:
            tree = self._tree

        return TreeStyler.create(tree=tree,
                                 initial_style=initial_style,
                                 mode=mode,
                                 highlight_style=highlight_style,
                                 node_style=node_style,
                                 legend_nsize=legend_nsize,
                                 legend_fsize=legend_fsize,
                                 annotate_color_present=annotate_color_present,
                                 annotate_color_absent=annotate_color_absent,
                                 annotate_color_unknown=annotate_color_unknown,
                                 annotate_opacity_present=annotate_opacity_present,
                                 annotate_opacity_absent=annotate_opacity_absent,
                                 annotate_opacity_unknown=annotate_opacity_unknown,
                                 annotate_border_color=annotate_border_color,
                                 annotate_kind=annotate_kind,
                                 annotate_box_width=annotate_box_width,
                                 annotate_box_height=annotate_box_height,
                                 annotate_border_width=annotate_border_width,
                                 annotate_margin=annotate_margin,
                                 annotate_guiding_lines=annotate_guiding_lines,
                                 annotate_guiding_lines_color=annotate_guiding_lines_color,
                                 figure_margin=figure_margin,
                                 show_border=show_border,
                                 title=title,
                                 title_fsize=title_fsize,
                                 legend_title=legend_title,
                                 annotate_show_box_label=annotate_show_box_label,
                                 annotate_box_label_color=annotate_box_label_color,
                                 annotate_arc_span=annotate_arc_span,
                                 annotate_label_fontsize=annotate_label_fontsize,
                                 include_unknown=include_unknown,
                                 show_leaf_names=show_leaf_names,
                                 leaf_name_fontsize=leaf_name_fontsize,
                                 leaf_name_func=leaf_name_func,
                                 sample_metadata=sample_metadata,
                                 show_legend_type_labels=show_legend_type_labels,
                                 legend_type_label_present=legend_type_label_present,
                                 legend_type_label_unknown=legend_type_label_unknown,
                                 rotation=rotation,
                                 allow_face_overlap=allow_face_overlap,
                                 show_branch_length=show_branch_length,
                                 show_branch_support=show_branch_support,
                                 tree_line_width=tree_line_width,
                                 tree_scale=tree_scale)

    @property
    def tree(self):
        return self._tree

    def has_tree(self) -> bool:
        return True



import logging
import tempfile
import time
from pathlib import Path
from typing import List, Tuple

import numpy as np
import pandas as pd
from sourmash.fig import load_matrix_and_labels

from genomics_data_index.storage.util import execute_commands

logger = logging.getLogger(__name__)


class KmerSearchManagerSourmash:

    def __init__(self, ncores: int = 1):
        self._ncores = ncores

    def search(self, kmer_size: int, similarity_threshold: float, query_file: Path,
               search_files: List[Path]) -> pd.DataFrame:
        start_time = time.time()
        logger.debug(f'Start search for matches to [{query_file}].')

        with tempfile.TemporaryDirectory() as tmp:
            out_file = Path(tmp, 'output.csv')
            command = ['sourmash', 'search', '-k', str(kmer_size), '--threshold', str(similarity_threshold),
                       '-o', str(out_file),
                       str(query_file)]
            command.extend([str(f) for f in search_files])

            execute_commands([command])
            end_time = time.time()
            logger.debug(f'Finished search for matches. Took {end_time - start_time:0.2f} seconds')

            return pd.read_csv(out_file)

    def distances(self, kmer_size: int, signature_files: List[Path]) -> Tuple[np.ndarray, List[str]]:
        sigs_to_file_threshold = 1000

        start_time = time.time()
        logger.debug(f'Start creating pairwise distance matrix.')

        with tempfile.TemporaryDirectory() as tmp:
            if len(signature_files) > sigs_to_file_threshold:
                signatures_list_file = Path(tmp, 'signatures')
                with open(signatures_list_file, 'w') as fh:
                    for file in signature_files:
                        fh.write(f'{file}\n')

                files_to_search = ['--from-file', str(signatures_list_file)]
            else:
                files_to_search = [str(f) for f in signature_files]

            out_file = Path(tmp, 'output-matrix')
            command = ['sourmash', 'compare', '-k', str(kmer_size), '--output', str(out_file), '-p', str(self._ncores)]
            command.extend(files_to_search)

            execute_commands([command])
            end_time = time.time()
            logger.debug(f'Finished search for matches. Took {end_time - start_time:0.2f} seconds')

            similarity_matrix, labels = load_matrix_and_labels(str(out_file))
            distance_matrix = 1 - similarity_matrix

            return distance_matrix, labels



from Bio import SeqIO

# Converts sequences in the reference fasta file to a genome BED file
# <https://bedtools.readthedocs.io/en/latest/content/general-usage.html#genome-file-format>
# Used to get the complement of all regions in the VCF file with reference/variants (i.e.,
# used to construct a BED file of all unknown/missing positions)
# Outputs a format like:
## SequenceID    length
# seq1  100
# seq2 500
# ...

input_fasta = snakemake.input.reference
output_file = snakemake.output.genome_bedtools

with open(output_file, 'w') as oh:
    with open(input_fasta, 'r') as fh:
        for record in SeqIO.parse(fh, 'fasta'):
            oh.write(f'{record.id}\t{len(record)}\n')



from genomics_data_index.storage.model.QueryFeatureMutationSPDI import QueryFeatureMutationSPDI


def test_create():
    f = QueryFeatureMutationSPDI('ref:1:A:T')
    assert 'ref:1:A:T' == f.id
    assert 'ref:1:A:T' == f.id_no_prefix


def test_to_unknown():
    f = QueryFeatureMutationSPDI('ref:1:A:T')
    assert f.to_unknown().id == 'ref:1:A:?'

    f = QueryFeatureMutationSPDI('ref:1:1:T')
    assert f.to_unknown().id == 'ref:1:1:?'

    f = QueryFeatureMutationSPDI('ref:10:AT:C')
    assert f.to_unknown().id == 'ref:10:AT:?'

    f = QueryFeatureMutationSPDI('ref:1:A:TT')
    assert f.to_unknown().id == 'ref:1:A:?'

    f = QueryFeatureMutationSPDI('ref:1:AT:TCG')
    assert f.to_unknown().id == 'ref:1:AT:?'

    f = QueryFeatureMutationSPDI('ref:1:2:TCG')
    assert f.to_unknown().id == 'ref:1:2:?'


def test_to_unknown_explode():
    f = QueryFeatureMutationSPDI('ref:1:A:T')
    assert [u.id for u in f.to_unknown_explode()] == ['ref:1:A:?']

    f = QueryFeatureMutationSPDI('ref:1:1:T')
    assert [u.id for u in f.to_unknown_explode()] == ['ref:1:1:?']

    f = QueryFeatureMutationSPDI('ref:10:AT:T')
    assert [u.id for u in f.to_unknown_explode()] == ['ref:10:A:?', 'ref:11:T:?']

    f = QueryFeatureMutationSPDI('ref:10:T:TC')
    assert [u.id for u in f.to_unknown_explode()] == ['ref:10:T:?']

    f = QueryFeatureMutationSPDI('ref:10:ATC:TCC')
    assert [u.id for u in f.to_unknown_explode()] == ['ref:10:A:?', 'ref:11:T:?', 'ref:12:C:?']

    f = QueryFeatureMutationSPDI('ref:10:3:TCC')
    assert [u.id for u in f.to_unknown_explode()] == ['ref:10:1:?', 'ref:11:1:?', 'ref:12:1:?']


def test_has_deletion_sequence():
    assert QueryFeatureMutationSPDI('ref:1:A:T').has_deletion_sequence()
    assert QueryFeatureMutationSPDI('ref:10:ATT:T').has_deletion_sequence()
    assert not QueryFeatureMutationSPDI('ref:10:1:T').has_deletion_sequence()
    assert not QueryFeatureMutationSPDI('ref:10:3:T').has_deletion_sequence()


def test_deletion_length():
    assert 1 == QueryFeatureMutationSPDI('ref:10:A:T').deletion_length()
    assert 3 == QueryFeatureMutationSPDI('ref:10:ATT:T').deletion_length()
    assert 1 == QueryFeatureMutationSPDI('ref:10:1:T').deletion_length()
    assert 3 == QueryFeatureMutationSPDI('ref:10:3:T').deletion_length()



import logging
from os import path

import yaml

logger = logging.getLogger(__name__)


def yaml_config_provider(file_path: str, cmd_name: str):
    if not path.exists(file_path):
        logger.warning(f'No configuration file found in [{file_path}]')
        return {}
    else:
        with open(file_path, 'r') as config_data:
            logger.debug(f'Reading config file {file_path}')
            data = yaml.safe_load(config_data)
            logger.debug(data)
            return data



import logging
import subprocess
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Tuple, List

from Bio import AlignIO
from Bio.Align import MultipleSeqAlignment
from Bio.Phylo.Applications import FastTreeCommandline
from ete3 import Tree

from genomics_data_index.storage.service import DatabaseConnection
from genomics_data_index.storage.service.CoreAlignmentService import CoreAlignmentService
from genomics_data_index.storage.service.ReferenceService import ReferenceService

logger = logging.getLogger(__name__)


class TreeService:
    TREE_BUILD_TYPES = ['fasttree', 'iqtree']
    ALIGN_TYPES = ['core', 'full']

    def __init__(self, database_connection: DatabaseConnection, reference_service: ReferenceService,
                 core_alignment_service: CoreAlignmentService):
        self._database = database_connection
        self._reference_service = reference_service
        self._core_alignment_service = core_alignment_service

    def build_tree(self, alignment: MultipleSeqAlignment,
                   tree_build_type: str = 'fasttree', num_cores: int = 1, align_type: str = 'core',
                   extra_params: str = None) -> Tuple[Tree, str]:
        if not tree_build_type in self.TREE_BUILD_TYPES:
            raise Exception(
                f'tree_type=[{tree_build_type}] is not one of the valid tree builder types {self.TREE_BUILD_TYPES}')

        if num_cores < 1:
            raise Exception(f'num_cores=[{num_cores}] is not supported')

        if extra_params is not None and tree_build_type == 'fasttree':
            raise Exception(f'extra_params is not supported for tree_build_type=[{tree_build_type}]')

        if align_type not in self.ALIGN_TYPES:
            raise Exception(f'align_type=[{align_type}] is not supported, must be one of {self.ALIGN_TYPES}')

        with TemporaryDirectory() as tmp_dir:
            input_file = Path(tmp_dir, 'input.fasta')
            with open(input_file, 'w') as f:
                AlignIO.write(alignment, f, 'fasta')

            if tree_build_type == 'fasttree':
                output_file = Path(tmp_dir, 'fasttree.tre')
                command = FastTreeCommandline(input=str(input_file), out=str(output_file))
                out, err = command()
                logger.debug('Output from FastTree')
                logger.debug(out)

                tree = Tree(str(output_file))
                return tree, out
            elif tree_build_type == 'iqtree':
                output_file = f'{str(input_file)}.treefile'
                command = ['iqtree', '--terrace', '--threads-max', str(num_cores), '-T', 'AUTO',
                           '-s', str(input_file)]

                extra_params_list = None
                if extra_params is not None:
                    extra_params_list = extra_params.split()
                    command.extend(extra_params_list)

                # Add ascertainment bias correction for core alignment
                # But only add if '-m' is not included in extra_params
                if extra_params_list is None or not ('-m' in extra_params_list):
                    if align_type == 'core':
                        command.extend(['-m', 'MFP+ASC'])
                    elif align_type == 'full':
                        command.extend(['-m', 'MFP'])

                try:
                    logger.debug(f'Running: {" ".join(command)}')
                    completed = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                               check=True, text=True)
                    out = str(completed.stdout)
                    tree = Tree(str(output_file))
                    return tree, out
                except subprocess.CalledProcessError as e:
                    err_msg = str(e.stderr.strip())
                    raise Exception(f'Could not run iqtree on alignment: error {err_msg}')
            else:
                raise Exception(f'tree_type=[{tree_build_type}] is invalid')

    def rebuild_tree(self, reference_name: str, num_cores: int = 1, tree_build_type='iqtree',
                     align_type='full', include_variants: List[str] = None, extra_params=None):
        logger.debug('Building alignment')
        alignment = self._core_alignment_service.construct_alignment(reference_name=reference_name,
                                                                     include_reference=True,
                                                                     include_variants=include_variants,
                                                                     align_type=align_type)

        logger.debug('Building tree')
        tree, out = self.build_tree(alignment=alignment,
                                    tree_build_type=tree_build_type,
                                    num_cores=num_cores,
                                    align_type=align_type,
                                    extra_params=extra_params)

        logger.debug(f'Updating tree for reference genome [{reference_name}]')
        self._reference_service.update_tree(reference_name=reference_name,
                                            tree=tree,
                                            alignment_length=alignment.get_alignment_length())



import math

from genomics_data_index.storage.index.KmerSearchManager import KmerSearchManagerSourmash
from genomics_data_index.test.integration import sourmash_dir

sigs = {
    'SampleA': sourmash_dir / 'SampleA.sig.gz',
    'SampleB': sourmash_dir / 'SampleB.sig.gz',
    'SampleC': sourmash_dir / 'SampleC.sig.gz',
}


def test_search_single_file():
    search_manager = KmerSearchManagerSourmash()

    results_df = search_manager.search(kmer_size=31, similarity_threshold=0.0,
                                       query_file=sigs['SampleA'], search_files=[sigs['SampleB'], sigs['SampleC']])
    assert 2 == len(results_df)
    assert ['SampleC', 'SampleB'] == list(results_df['name'].tolist())
    assert math.isclose(0.5, results_df['similarity'].tolist()[0], rel_tol=1e-3)
    assert math.isclose(0.478, results_df['similarity'].tolist()[1], rel_tol=1e-3)


def test_search_single_file_higher_threshold():
    search_manager = KmerSearchManagerSourmash()

    results_df = search_manager.search(kmer_size=31, similarity_threshold=0.49,
                                       query_file=sigs['SampleA'], search_files=[sigs['SampleB'], sigs['SampleC']])
    assert 1 == len(results_df)
    assert ['SampleC'] == list(results_df['name'].tolist())
    assert math.isclose(0.5, results_df['similarity'].tolist()[0], rel_tol=1e-3)


def test_search_single_file_higher_threshold_no_results():
    search_manager = KmerSearchManagerSourmash()

    results_df = search_manager.search(kmer_size=31, similarity_threshold=0.9,
                                       query_file=sigs['SampleA'], search_files=[sigs['SampleB'], sigs['SampleC']])
    assert 0 == len(results_df)


def test_search_single_file_2():
    search_manager = KmerSearchManagerSourmash()

    results_df = search_manager.search(kmer_size=31, similarity_threshold=0.0,
                                       query_file=sigs['SampleB'], search_files=[sigs['SampleA'], sigs['SampleC']])
    assert 2 == len(results_df)
    assert ['SampleC', 'SampleA'] == list(results_df['name'].tolist())
    assert math.isclose(0.681, results_df['similarity'].tolist()[0], rel_tol=1e-3)
    assert math.isclose(0.478, results_df['similarity'].tolist()[1], rel_tol=1e-3)


def test_distances_all():
    search_manager = KmerSearchManagerSourmash()

    results_d, labels = search_manager.distances(kmer_size=31,
                                                 signature_files=[sigs['SampleA'], sigs['SampleB'], sigs['SampleC']])
    assert (3, 3) == results_d.shape
    assert ['SampleA', 'SampleB', 'SampleC'] == labels

    l = {element: idx for idx, element in enumerate(labels)}

    assert math.isclose(results_d[l['SampleA']][l['SampleA']], 0, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleA']][l['SampleB']], 0.522, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleA']][l['SampleC']], 0.5, rel_tol=1e-3)

    assert math.isclose(results_d[l['SampleB']][l['SampleA']], 0.522, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleB']][l['SampleB']], 0, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleB']][l['SampleC']], 0.3186, rel_tol=1e-3)

    assert math.isclose(results_d[l['SampleC']][l['SampleA']], 0.5, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleC']][l['SampleB']], 0.3186, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleC']][l['SampleC']], 0, rel_tol=1e-3)


def test_distances_pair():
    search_manager = KmerSearchManagerSourmash()

    results_d, labels = search_manager.distances(kmer_size=31,
                                                 signature_files=[sigs['SampleA'], sigs['SampleC']])
    assert (2, 2) == results_d.shape
    assert ['SampleA', 'SampleC'] == labels

    l = {element: idx for idx, element in enumerate(labels)}

    assert math.isclose(results_d[l['SampleA']][l['SampleA']], 0, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleA']][l['SampleC']], 0.5, rel_tol=1e-3)

    assert math.isclose(results_d[l['SampleC']][l['SampleA']], 0.5, rel_tol=1e-3)
    assert math.isclose(results_d[l['SampleC']][l['SampleC']], 0, rel_tol=1e-3)


def test_distances_single():
    search_manager = KmerSearchManagerSourmash()

    results_d, labels = search_manager.distances(kmer_size=31,
                                                 signature_files=[sigs['SampleA']])
    assert (1, 1) == results_d.shape
    assert ['SampleA'] == labels

    l = {element: idx for idx, element in enumerate(labels)}

    assert math.isclose(results_d[l['SampleA']][l['SampleA']], 0, rel_tol=1e-3)



import pandas as pd
import pytest

from genomics_data_index.storage.io.mlst.MLSTTSeemannFeaturesReader import MLSTTSeemannFeaturesReader
from genomics_data_index.test.integration import basic_mlst_file, mlst_file_unknown, expand_list_by


@pytest.fixture
def mlst_reader() -> MLSTTSeemannFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=basic_mlst_file)


@pytest.fixture
def mlst_reader_unknown() -> MLSTTSeemannFeaturesReader:
    return MLSTTSeemannFeaturesReader(mlst_file=mlst_file_unknown)


def test_get_features_table(mlst_reader):
    num_samples = 6
    num_loci = 7

    mlst_df = mlst_reader.get_features_table()

    assert ['File', 'Sample', 'Scheme', 'Locus', 'Allele', 'Sequence Type'] == list(mlst_df.columns)

    assert num_samples * num_loci == len(mlst_df)
    assert expand_list_by(['2014C-3598.fasta', '2014C-3599.fasta', '2014D-0067.fasta', '2014D-0068.fasta',
                           'CFSAN002349.fasta', 'CFSAN023463.fasta'], num_loci) == list(mlst_df['File'].tolist())
    assert expand_list_by(['2014C-3598', '2014C-3599', '2014D-0067', '2014D-0068',
                           'CFSAN002349', 'CFSAN023463'], num_loci) == list(mlst_df['Sample'].tolist())

    assert ['abcZ', 'bglA', 'cat', 'dapE', 'dat', 'ldh', 'lhkA'] == list(mlst_df.loc[mlst_df['Sample'] == 'CFSAN002349',
                                                                                     'Locus'].tolist())
    assert ['1', '51', '11', '13', '2', '5', '4'] == list(mlst_df.loc[mlst_df['Sample'] == 'CFSAN002349',
                                                                      'Allele'].tolist())
    assert ['adk', 'fumC', 'gyrB', 'icd', 'mdh', 'purA', 'recA'] == list(mlst_df.loc[mlst_df['Sample'] == '2014C-3598',
                                                                                     'Locus'].tolist())
    assert ['100', '23', '68', '45', '1', '35', '7'] == list(mlst_df.loc[mlst_df['Sample'] == '2014C-3598',
                                                                         'Allele'].tolist())


def test_get_features_table_with_unknown(mlst_reader_unknown):
    num_samples = 4
    num_loci = 7

    mlst_df = mlst_reader_unknown.get_features_table()

    assert ['File', 'Sample', 'Scheme', 'Locus', 'Allele', 'Sequence Type'] == list(mlst_df.columns)

    assert num_samples * num_loci == len(mlst_df)

    assert ['abcZ', 'bglA', 'cat', 'dapE', 'dat', 'ldh', 'lhkA'] == list(mlst_df.loc[mlst_df['Sample'] == 'CFSAN002349',
                                                                                     'Locus'].tolist())
    assert ['1', '?', '11', '13', '2', '5', '5'] == list(mlst_df.loc[mlst_df['Sample'] == 'CFSAN002349',
                                                                     'Allele'].tolist())
    assert ['abcZ', 'bglA', 'cat', 'dapE', 'dat', 'ldh', 'lhkA'] == list(mlst_df.loc[mlst_df['Sample'] == 'CFSAN023463',
                                                                                     'Locus'].tolist())
    assert ['?', '?', '11', '13', '2', '5', '5'] == list(mlst_df.loc[mlst_df['Sample'] == 'CFSAN023463',
                                                                     'Allele'].tolist())
    assert ['adk', 'fumC', 'gyrB', 'icd', 'mdh', 'purA', 'recA'] == list(mlst_df.loc[mlst_df['Sample'] == '2014C-3598',
                                                                                     'Locus'].tolist())
    assert ['100', '?', '?', '45', '1', '35', '7'] == list(mlst_df.loc[mlst_df['Sample'] == '2014C-3598',
                                                                       'Allele'].tolist())


def test_iter_data(mlst_reader: MLSTTSeemannFeaturesReader):
    sample_data_map = {}
    for sample_name, scheme, allele_data in mlst_reader.iter_sample_data():
        if sample_name in sample_data_map:
            raise Exception(f'Duplicate sample {sample_name} in sample_data_map')
        sample_data_map[sample_name] = {'scheme': scheme, 'allele_data': allele_data}

    assert 6 == len(sample_data_map)

    sample_schemes = {(n, sample_data_map[n]['scheme']) for n in sample_data_map}
    assert {('2014C-3598', 'ecoli'), ('2014C-3599', 'ecoli'),
            ('2014D-0067', 'campylobacter'), ('2014D-0068', 'campylobacter'),
            ('CFSAN002349', 'lmonocytogenes'), ('CFSAN023463', 'lmonocytogenes')} == sample_schemes

    allele_data = pd.Series({'adk': '100', 'fumC': '23', 'gyrB': '68', 'icd': '45',
                             'mdh': '1', 'purA': '35', 'recA': '7'}, name='Allele')
    allele_data.index.name = 'Locus'
    assert allele_data.equals(sample_data_map['2014C-3598']['allele_data'])

    allele_data = pd.Series({'abcZ': '1', 'bglA': '51', 'cat': '11', 'dapE': '13',
                             'dat': '2', 'ldh': '5', 'lhkA': '5'}, name='Allele')
    allele_data.index.name = 'Locus'
    assert allele_data.equals(sample_data_map['CFSAN023463']['allele_data'])


def test_samples_list(mlst_reader):
    assert {'2014C-3598', '2014C-3599', '2014D-0067', '2014D-0068',
            'CFSAN002349', 'CFSAN023463'} == set(mlst_reader.samples_list())



from __future__ import annotations

from typing import Union, Any, Optional, Callable, Dict

import pandas as pd

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.cluster.ClusterScoreMRCAJaccard import ClusterScoreMRCAJaccard
from genomics_data_index.api.query.impl.cluster.ClusterScoreMethod import ClusterScoreMethod
from genomics_data_index.storage.SampleSet import SampleSet


class ClusterScorer:
    """
    A ClusterScorer is used to assign scores to clusters of samples based on a tree.
    """

    SCORE_KINDS: Dict[str, ClusterScoreMethod] = {
        'mrca_jaccard': ClusterScoreMRCAJaccard()
    }

    def __init__(self, universe_samples: SamplesQuery):
        """
        Builds a new ClusterScorer with the given universe samples.
        :param universe_samples: The universe of samples used for scoring.
        :return: A new ClusterScorer.
        """
        self._universe_samples = universe_samples

    def score_samples(self, samples: Union[SamplesQuery, SampleSet], kind: str = 'mrca_jaccard') -> float:
        """
        Gives a score for how well the passed set of samples is clustered together in a tree.

        :param samples: The samples to score.
        :param kind: The kind of scoring method to use.
        :return: A score for how well the passed set of samples is clustered together in a tree.
        """
        if kind in self.SCORE_KINDS:
            return self.SCORE_KINDS[kind].score(samples, self._universe_samples)
        else:
            raise Exception(f'kind=[{kind}] is invalid. Must be one of {self.SCORE_KINDS}')

    def score_groupby(self, groupby_column: str, kind: str = 'mrca_jaccard',
                      min_samples_count: Optional[int] = None, max_samples_count: Optional[int] = None,
                      groupby_func: Callable[[Any], str] = None,
                      na_value: Any = None) -> pd.DataFrame:
        """
        Gives a score for how well sets of samples defined by the groupby_column cluster together in a tree.

        :param groupby_column: The column in the samples query dataframe to group by. When using this option you likely
                               will want to join your query with an external data frame defining extra metadata associated
                               with the samples.
        :param kind: The kind of scoring method to use.
        :param min_samples_count: The minimum samples in a group to include for scoring (default no minimum).
        :param max_samples_count: The maximum samples in a group to include for scoring (default no maximum).
        :param groupby_func: An optional function to apply to the groupby_column to be used to define the groups.
                             This gets passed a particular value from the groupby_column and returns a string representing
                             the group this value belongs to. For example if one value in 'groupby_column' is A.B.C then
                             groupby_func('A.B.C') => 'A' would say this particular row belongs to group 'A'.
        :param na_value: The value used to replace pd.NA with. If this is None (default) then NA values are excluded.
        :return: A dataframe containing scores and counts of samples (indexed by the groups in the groupby_column)
         for how well the passed set of samples is clustered together in a tree.
        """
        if kind not in self.SCORE_KINDS:
            raise Exception(f'kind=[{kind}] is invalid. Must be one of {self.SCORE_KINDS}')
        else:
            universe_df = self._universe_samples.toframe()

            if groupby_column not in universe_df:
                raise Exception(
                    f'column={groupby_column} not found in data frame from query={self._universe_samples}. '
                    f'Perhaps you meant to join this query with a dataframe before scoring clusters?')

            universe_sub_columns = universe_df[['Sample ID', groupby_column]]
            if na_value is not None:
                universe_sub_columns = universe_sub_columns.fillna(na_value)

            if groupby_func is None:
                groups_sample_sets_df = universe_sub_columns.groupby(groupby_column).agg(SampleSet)
            else:
                groups_sample_sets_df = universe_sub_columns.set_index(groupby_column).groupby(by=groupby_func).agg(
                    SampleSet)
                groups_sample_sets_df.index.name = groupby_column

            groups_sample_sets_df['Sample Count'] = groups_sample_sets_df.apply(
                lambda x: len(x['Sample ID']), axis='columns')

            # Subset before scoring to not waste time scoring groups we don't need
            if min_samples_count is not None:
                groups_sample_sets_df = groups_sample_sets_df[
                    groups_sample_sets_df['Sample Count'] >= min_samples_count]
            if max_samples_count is not None:
                groups_sample_sets_df = groups_sample_sets_df[
                    groups_sample_sets_df['Sample Count'] <= max_samples_count]

            # Do scoring
            groups_sample_sets_df['Score'] = groups_sample_sets_df.apply(
                lambda x: self.score_samples(x['Sample ID']), axis='columns')

            return groups_sample_sets_df[['Score', 'Sample Count']]

    @classmethod
    def register_score_kind(cls, name: str, score_kind: ClusterScoreMethod) -> None:
        """
        Registers a new ClusterScoreMethod. This can be used to dynamically add new scoring implementations to
        the ClusterScorer.
        :param name: The name of the ClusterScoreMethod. This will be used as the 'kind' parameter in ClusterScorder.score_samples().
        :param score_kind: The ClusterScoreMethod implementation.
        :return: Nothing. Maps the name to the ClusterScoreMethod implementation.
        """
        if name not in cls.SCORE_KINDS:
            cls.SCORE_KINDS[name] = score_kind
        else:
            raise Exception(f'Score kind with name={name} already exists {cls.SCORE_KINDS[name]}')

    @classmethod
    def create(cls, universe_samples: SamplesQuery) -> ClusterScorer:
        """
        Creates a new ClusterScorer used to score sets of samples in a tree based on how well they cluster together.

        :param universe_samples: The set of samples under consideration. This should have a tree attached to it
                                 (i.e., should be of type TreeSamplesQuery).
        :return: A new ClusterScorer for the passed samples.
        """
        if not universe_samples.has_tree():
            raise Exception(f'Passed universe_samples={universe_samples} does not have an associated tree. '
                            f'Perhaps you forgot to join a tree with join_tree() or build with build_tree()?')

        return ClusterScorer(universe_samples=universe_samples)



from __future__ import annotations

import logging
from typing import cast, Optional, Set

from ete3 import Tree, ClusterTree

from genomics_data_index.api.query.SamplesQuery import SamplesQuery
from genomics_data_index.api.query.impl.ExperimentalTreeSamplesQuery import ExperimentalTreeSamplesQuery
from genomics_data_index.api.query.impl.KmerTreeSamplesQuery import KmerTreeSamplesQuery
from genomics_data_index.api.query.impl.MutationTreeSamplesQuery import MutationTreeSamplesQuery
from genomics_data_index.api.query.impl.TreeBuilderKmers import TreeBuilderKmers
from genomics_data_index.api.query.impl.TreeBuilderReferenceMutations import TreeBuilderReferenceMutations
from genomics_data_index.api.query.impl.TreeSamplesQuery import TreeSamplesQuery
from genomics_data_index.configuration.connector.DataIndexConnection import DataIndexConnection
from genomics_data_index.storage.SampleSet import SampleSet
from genomics_data_index.storage.model.db import Reference

logger = logging.getLogger(__name__)


class TreeSamplesQueryFactory:
    """
    Builds TreeSamplesQuery objects.
    Used mainly to switch between regular and experimental tree queries while avoiding circular dependencies.
    """
    BUILD_TREE_KINDS = ['mutation', 'mutations', 'mutation_experimental', 'mutations_experimental',
                        'kmer', 'kmers']

    _instance = None

    def __init__(self):
        pass

    def _prune_tree_to_query_samples(self, tree: Tree,
                                     leaf_names: Set[str],
                                     leaf_names_in_query: Set[str],
                                     reference_name: Optional[str]) -> Tree:
        include_reference = reference_name is not None
        leaves_missing_in_database = leaf_names - leaf_names_in_query

        if len(leaves_missing_in_database) == 0:
            if include_reference:
                logger.warning(f'Reference genome name [{reference_name}] is the same as a sample name already '
                               f'in the database.')
        else:
            if not include_reference or leaves_missing_in_database != {reference_name}:
                if include_reference:
                    leaves_to_prune = leaf_names_in_query.copy()
                    leaves_to_prune.add(reference_name)

                    leaves_missing_not_reference = leaves_missing_in_database.copy()

                    if reference_name in leaves_missing_not_reference:
                        leaves_missing_not_reference.remove(reference_name)

                    extra_msg = ' (and not the reference genome name)'
                else:
                    leaves_to_prune = leaf_names_in_query
                    leaves_missing_not_reference = leaves_missing_in_database

                    extra_msg = ''

                if len(leaves_missing_not_reference) > 5:
                    missing_str = ', '.join(list(leaves_missing_not_reference)[0:5])
                    missing_str += ', ...'
                else:
                    missing_str = ', '.join(list(leaves_missing_not_reference))

                logger.warning(
                    f'{len(leaves_missing_not_reference)}/{len(leaf_names)} leaves in the tree are'
                    f' not found in the query{extra_msg}: [{missing_str}].'
                    f' Pruning tree to contain only those samples in the query ('
                    f'{len(leaves_to_prune)}/{len(leaf_names)}).')

                tree = tree.copy(method='newick')
                tree.prune(leaves_to_prune, preserve_branch_length=True)

        return tree

    def _join_tree_mutations(self, tree: Tree, kind: str, database_connection: DataIndexConnection,
                             wrapped_query: SamplesQuery,
                             leaf_names: Set[str],
                             leaf_names_in_query: Set[str],
                             alignment_length: int,
                             reference_name: str = None) -> TreeSamplesQuery:
        if alignment_length is None:
            raise Exception(f'Must set alignment_length for mutation tree.')
        else:
            tree = self._prune_tree_to_query_samples(tree=tree,
                                                     leaf_names=leaf_names,
                                                     leaf_names_in_query=leaf_names_in_query,
                                                     reference_name=reference_name)

            return self._create_tree_samples_query_from_tree(kind=kind,
                                                             connection=database_connection,
                                                             wrapped_query=wrapped_query,
                                                             tree=tree,
                                                             alignment_length=alignment_length,
                                                             reference_name=reference_name,
                                                             include_reference=reference_name is not None)

    def _join_tree_kmers(self, tree: Tree, kind: str, database_connection: DataIndexConnection,
                         wrapped_query: SamplesQuery,
                         leaf_names: Set[str],
                         leaf_names_in_query: Set[str]) -> TreeSamplesQuery:
        tree = self._prune_tree_to_query_samples(tree=tree,
                                                 leaf_names=leaf_names,
                                                 leaf_names_in_query=leaf_names_in_query,
                                                 reference_name=None)

        if not isinstance(tree, ClusterTree):
            logger.warning(f'Passed tree={tree} is not a {ClusterTree.__name__}. '
                           f'Converting to a {ClusterTree.__name__}')
            newick = tree.write()
            cluster_tree = ClusterTree(newick=newick)
        else:
            cluster_tree = cast(ClusterTree, tree)

        return self._create_tree_samples_query_from_tree(kind=kind,
                                                         connection=database_connection,
                                                         wrapped_query=wrapped_query,
                                                         tree=cluster_tree,
                                                         alignment_length=None,
                                                         reference_name=None,
                                                         include_reference=False)

    def join_tree(self, tree: Tree, kind: str, database_connection: DataIndexConnection,
                  wrapped_query: SamplesQuery, **kwargs) -> TreeSamplesQuery:
        if kind not in self.BUILD_TREE_KINDS:
            raise Exception(f'Invalid kind=[{kind}]. Must be one of {self.BUILD_TREE_KINDS}.')

        tree = tree.copy(method='newick')

        leaf_names = set(tree.get_leaf_names())
        sample_name_ids = database_connection.sample_service.find_sample_name_ids(leaf_names)
        samples_set = SampleSet(sample_name_ids.values())

        wrapped_query_tree_set = wrapped_query.intersect(sample_set=samples_set,
                                                         query_message=f'join_tree({len(leaf_names)} leaves)')

        leaf_names_in_query = wrapped_query_tree_set.toset(names=True)

        if kind == 'mutation' or kind == 'mutation_experimental':
            return self._join_tree_mutations(tree=tree, kind=kind, database_connection=database_connection,
                                             wrapped_query=wrapped_query_tree_set,
                                             leaf_names=leaf_names,
                                             leaf_names_in_query=leaf_names_in_query,
                                             **kwargs)
        elif kind == 'kmer' or kind == 'kmers':
            return self._join_tree_kmers(tree=tree, kind=kind, database_connection=database_connection,
                                         wrapped_query=wrapped_query_tree_set,
                                         leaf_names=leaf_names,
                                         leaf_names_in_query=leaf_names_in_query)
        else:
            raise Exception(f'Invalid kind=[{kind}]. Must be one of {self.BUILD_TREE_KINDS}.')

    def build_tree(self, kind: str, database_connection: DataIndexConnection,
                   wrapped_query: SamplesQuery, scope: str = None, include_reference: bool = True,
                   **kwargs) -> TreeSamplesQuery:
        if kind == 'mutation' or kind == 'mutation_experimental':
            if scope is None:
                raise Exception(f'Invalid scope=[{scope}]. You must specify a scope (i.e., reference genome name) for '
                                f'kind=[{kind}].')

            tree_builder = TreeBuilderReferenceMutations(database_connection,
                                                         reference_name=scope)
            tree, alignment_length, tree_samples_set = tree_builder.build(wrapped_query.sample_set,
                                                                          include_reference=include_reference,
                                                                          **kwargs)
        elif kind == 'kmer' or kind == 'kmers':
            tree_builder = TreeBuilderKmers(database_connection)
            tree, alignment_length, tree_samples_set = tree_builder.build(wrapped_query.sample_set,
                                                                          **kwargs)
        else:
            raise Exception(f'Got kind=[{kind}], only the following kinds are supported: {self.BUILD_TREE_KINDS}')

        wrapped_query_tree_set = wrapped_query.intersect(sample_set=tree_samples_set,
                                                         query_message=f'mutation_tree({scope})')

        return self._create_tree_samples_query_from_tree(kind=kind,
                                                         connection=database_connection,
                                                         wrapped_query=wrapped_query_tree_set,
                                                         tree=tree,
                                                         alignment_length=alignment_length,
                                                         reference_name=scope,
                                                         include_reference=include_reference)

    def _create_tree_samples_query_from_tree(self, kind: str, tree: Tree,
                                             alignment_length: Optional[int],
                                             reference_name: Optional[str],
                                             connection: DataIndexConnection,
                                             wrapped_query: SamplesQuery,
                                             include_reference: bool = True) -> TreeSamplesQuery:
        if kind == 'mutation' or kind == 'mutations':
            return MutationTreeSamplesQuery(connection=connection,
                                            wrapped_query=wrapped_query,
                                            tree=tree,
                                            alignment_length=alignment_length,
                                            reference_name=reference_name,
                                            reference_included=include_reference)
        elif kind == 'mutation_experimental' or kind == 'mutations_experimental':
            return ExperimentalTreeSamplesQuery(connection=connection,
                                                wrapped_query=wrapped_query,
                                                tree=tree,
                                                universe_tree=tree,
                                                alignment_length=alignment_length,
                                                reference_name=reference_name,
                                                reference_included=include_reference)
        elif kind == 'kmer' or kind == 'kmers':
            if isinstance(tree, ClusterTree):
                tree = cast(ClusterTree, tree)
            else:
                raise Exception(
                    f'Invalid type for tree=[{tree}]. Expected [{ClusterTree.__name__}], got [{type(tree)}]')

            return KmerTreeSamplesQuery(connection=connection,
                                        wrapped_query=wrapped_query,
                                        tree=tree)
        else:
            raise Exception(f'Got kind=[{kind}], only the following kinds are supported: {self.BUILD_TREE_KINDS}')

    def create_from_reference_genome(self, kind: str, reference_genome: Reference,
                                     connection: DataIndexConnection,
                                     wrapped_query: SamplesQuery,
                                     include_reference: bool = True):
        return self._create_tree_samples_query_from_tree(kind=kind,
                                                         connection=connection, wrapped_query=wrapped_query,
                                                         tree=reference_genome.tree,
                                                         alignment_length=reference_genome.tree_alignment_length,
                                                         reference_name=reference_genome.name,
                                                         include_reference=include_reference)

    @classmethod
    def instance(cls) -> TreeSamplesQueryFactory:
        if cls._instance is None:
            cls._instance = TreeSamplesQueryFactory()
        return cls._instance
